/*
"The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the content of this file.

The Initial Developer of the Original Code is Dr. Robert Kofler.
Carrer del Baluard; 73 21a
08003 Barcelona
All Rights Reserved.

Contributor(s): Dr. Robert Kofler.

Alternatively, the contents of this file may be used under the terms
of the _____ license (the  "[___] License"), in which case the
provisions of [______] License are applicable instead of those
above. If you wish to allow use of your version of this file only
under the terms of the [____] License and not to allow others to use
your version of this file under the MPL, indicate your decision by
deleting the provisions above and replace them with the notice and
other provisions required by the [___] License. If you do not delete
the provisions above, a recipient may use your version of this file
under either the MPL or the [___] License."
*/
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Collections;
using Bio.Seq;
using Bio.SNP;
using Bio.Alignment.Pairwise;

namespace Bio
{
    namespace IO
    {
        namespace SNP
        {
            using Bio.SNP;

            /// <summary>
            /// Reads SSRs from a TabDelimitedSSR-File
            /// </summary>
            public class RefSNPReader 
            {
                private StreamReader sr;
                private readonly int bufferSize;
                public static int linecount=0;



                public RefSNPReader(StreamReader sr, int bufferSize)
                {
                    this.sr = sr;
                    this.bufferSize = bufferSize;

                }
                public RefSNPReader(StreamReader sr)
                {
                    this.sr = sr;
                    this.bufferSize = 5000;

                }

                public List<RefSNP> ReadRefSNPs()
                {
                    List<RefSNP> refSNPList = new List<RefSNP>();
                    string line = "";


                    //sw.WriteLine("Seq_Name\tMotif\tMotif_Standardised\tSSR_Start\tSSR_End\tLength\tScore\tMismatches\tSequence"); //9

#if DEBUG
                    System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                    stopw.Start();
#endif
                    //As long as there are lines in rhe file proceed!
                    while ((refSNPList.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                    {
                         linecount++;
                        //Get rid of space
                        line = line.Trim();
                        if (line == "" || linecount < 6) continue;
                        else
                        {
                            refSNPList.Add(ParseTabDelimitedRefSNP(line));

                        }
                    }
#if DEBUG
                    stopw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRReaderTabDelimited; {1} SSRs read in {2} ms", DateTime.Now.TimeOfDay, refSNPList.Count, stopw.ElapsedMilliseconds));
#endif

                    return refSNPList;
                }





                private RefSNP ParseTabDelimitedRefSNP(string toParse)
                {
                    /*
 * Column   Data
  1      RefSNP id (rs#)
  2      mapweight where
            1 = Unmapped
            2 = Mapped to single position in genome
            3 = Mapped to 2 positions on a single chromosome
            4 = Mapped to 3-10 positions in genome (possible paralog hits)
            5 = Mapped to >10 positions in genome
  3      snp_type where
            0 = Not withdrawn.
            1 = Withdrawn. There are several reasons for withdrawn, the
                withdrawn status is fully defined in the asn1, flatfile,
                and XML descriptions of the RefSNP. See /specs/docsum_2005.asn
                for a full definition of snp-type values.
  4      Total number of chromosomes hit by this RefSNP during mapping
  5      Total number of contigs hit by this RefSNP during mapping
  6      Total number of hits to genome by this RefSNP during mapping
  7      Chromosome for this hit to genome
  8      Contig accession for this hit to genome
  9      Version number of contig accession for this hit to genome
 10      Contig ID for this hit to genome
 11      Position of RefSNP in contig coordinates
 12      Position of RefSNP in chromosome coordinates (used to order report)
            Locations are specified in NCBI sequence location convention where:
                   x, a single number, indicates a feature at base position x
                   x..y, denotes a feature that spans from x to y inclusive.
                   x^y, denotes a feature that is inserted between bases x and y
 13      Genes at this same position on the chromosome
 14      Average heterozygosity of this RefSNP
 15      Standard error of average heterozygosity
 16      Maximum reported probability that RefSNP is real. (For computationally-
             predicted submissions)
 17      Validated status
             0 = No validation information
             1 = Cluster has 2+ submissions, with 1+ submission assayed 
                 with a non-computational method
             2 = At least one subsnp in cluster has frequency data submitted
             3 = Non-computational method in cluster and frequency data present
             4 = At lease one subsnp in cluster has been experimentally 
                 validated by submitter
                 for other validation status values, please see:
                 <a href="ftp://ftp.ncbi.nih.gov/snp/database/organism_shared_data/SnpValidationCode.bcp.gz">ftp://ftp.ncbi.nih.gov/snp/database/organism_shared_data
                 /SnpValidationCode.bcp.gz</a>
 18      Genotypes available in dbSNP for this RefSNP
             1 = yes
             0 = no
 19      Linkout available to submitter website for further data on the RefSNP
             1 = yes
             0 = no
 20      dbSNP build ID when the refSNP was first created (i.e. the create date)
 21      dbSNP build ID of the most recent change to the refSNP cluster. The
         date of the change is represented by the build ID which has an
         approximate date/time associated with it. (see:
         http://www.ncbi.nlm.nih.gov/projects/SNP/buildhistory.cgi)
22       Mapped to a reference or alternate (e.g. Celera) assembly 
 */
                    string[] division = toParse.Split(new char[] { '\t' });
                    if (division.Length != 22) throw new Exception("Error during reading of Tab Delimited RefSNP-File.");
                    division[10] = division[10].Trim();
                    division[11] = division[11].Trim();
                    division[13] = division[13].Trim();
                    division[14] = division[14].Trim();
                    division[15] = division[15].Trim();
                    System.Globalization.NumberFormatInfo ni = new System.Globalization.NumberFormatInfo();
                    ni.NumberDecimalSeparator=".";
#if DEBUG
                    Convert.ToInt32(division[0]);
                    Convert.ToInt32(division[1]); //int mapweight
                    Convert.ToInt32(division[2]); //int snpType
                    Convert.ToInt32(division[3]); //int chromosomeHits
                    Convert.ToInt32(division[4]); //int contigHits
                    Convert.ToInt32(division[5]); //int genomeHits
                   string temp= division[6]                 ; //string chromosome
                    temp=division[7]                 ; //string contigAccession
                    Convert.ToInt32(division[8]); //int contigVersionNumber
                    temp=division[9]                 ; //string contigID
                    int? testint=division[10]==""?(int?)null:Convert.ToInt32(division[10]);//int startPostionInContig
                    testint=division[11]==""?(int?)null:Convert.ToInt32(division[11]);//int startPostionInChromosome
                    string[] temp2=division[12].Contains(",")?division[12].Split(new char[] {',' }):null;//string[] genes
                    double? av= division[13]==""?(double?)null:Double.Parse(division[13],ni); //double? averageHeterozygosity
                    av= division[14] == "" ? (double?)null : Double.Parse(division[14],ni); //double? stdDev_averageHeterozygosity
                    av= (division[15] == "") ? (double?)null : Double.Parse(division[15],ni); //double? p_realRefSnP
                    Convert.ToInt32(division[16]);     //int validationStatus
                    bool test=division[17]=="1"?true:false;      //bool genotypesAvailable
                    test=division[18]=="1"?true:false;      //bool linkOutAvailable
                    Convert.ToInt32(division[19]);      //int dbSNPBuildFirtsBuild
                    Convert.ToInt32(division[20]);     //int dbSNPBuildRecentChange
                    temp=division[21]; 
#endif



                    //Add the Microsatellite
                    return new RefSNP(Convert.ToInt32(division[0]) //int RefSNPId
                                     , Convert.ToInt32(division[1]) //int mapweight
                                     , Convert.ToInt32(division[2]) //int snpType
                                     , Convert.ToInt32(division[3]) //int chromosomeHits
                                     , Convert.ToInt32(division[4]) //int contigHits
                                     , Convert.ToInt32(division[5]) //int genomeHits
                                     , division[6]                  //string chromosome
                                     , division[7]                  //string contigAccession
                                     , Convert.ToInt32(division[8]) //int contigVersionNumber
                                     , division[9]                  //string contigID
                                     , division[10]==""?(int?)null:Convert.ToInt32(division[10])//int startPostionInContig
                                     , division[11]==""?(int?)null:Convert.ToInt32(division[11])//int startPostionInChromosome
                                     ,division[12].Contains(",")?division[12].Split(new char[] {',' }):null//string[] genes
                                     ,division[13]==""?(double?)null:Double.Parse(division[13],ni) //double? averageHeterozygosity
                                     , division[14] == "" ? (double?)null : Double.Parse(division[14],ni) //double? stdDev_averageHeterozygosity
                                     , division[15] == "" ? (double?)null : Double.Parse(division[15],ni) //double? p_realRefSnP
                                     ,Convert.ToInt32(division[16])     //int validationStatus
                                     ,division[17]=="1"?true:false      //bool genotypesAvailable
                                     ,division[18]=="1"?true:false      //bool linkOutAvailable
                                     ,Convert.ToInt32(division[19])      //int dbSNPBuildFirtsBuild
                                     ,Convert.ToInt32(division[20])     //int dbSNPBuildRecentChange
                                     ,division[21]                      //string mappedToReference
                                     );

                }

            }


            /// <summary>
            /// Instance of a SNP filewriter; SNPs are written in the tab delimited format
            /// </summary>
            public class SNPWriter 
            {
                private StreamWriter sw;
                private bool wroteIni = false;
                private bool wroteGeneralInfo = false;
                public SNPWriter(StreamWriter sw)
                {
                    this.sw = sw;

                }
                public void WriteSNPs(List<SNP> snps)
                {

#if DEBUG
                    System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                    stopw.Start();
#endif
                    if (!wroteIni) //tdatabase_name\tquery_name\tstrand\tdatabase_char\tquery_char\tposition_database\tposition_query\tallele_identification\tinvalid_allele\tallele_id\tsynonymous")

                    {
                        sw.WriteLine("#");
                        sw.WriteLine("#\tdatabase_name\tquery_name\tstrand\tdatabase_char\tquery_char\tsiteID_database\tsiteID_query\tvalid\tdistance_from_alignment_ends\talignment_Quality_Token\tsequence_quality_at_site\tsequence_quality_in_neighborhood\tallele_id");
                        sw.WriteLine("#");
                        sw.WriteLine();
                        wroteIni = true;
                    }

                    if (!wroteGeneralInfo && snps.Count > 0)
                    {

                        sw.WriteLine(String.Format(";Mode {0}",snps[0].IsFullSNP==true?"fullSNP":"halfSNP"));
                        sw.WriteLine();
                        wroteGeneralInfo = true;
                    }

                    foreach (SNP s in snps)
                    {
                        sw.WriteLine(String.Format("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}\t{12}"
                            ,s.DatabaseParent.Replace('\t',' ')
                            ,s.QueryParent.Replace('\t',' ')
                            ,s.PlusPlusStrand==true?"+/+":"+/-"
                            ,s.DatabaseChar
                            ,s.QueryChar
                            ,s.SiteID_Database
                            ,s.SiteID_Query
                            ,s.Valid.ToString()
                            ,s.DistanceFromAlignmentEnd
                            ,s.AlignmentQualityTokens
                            ,s.SequenceQualityAtSite
                            ,s.SequenceQualityInNeighborhood
                            ,s.AlleleID!=null?s.AlleleID.Replace('\t',' '):""
                            ));
                    }

#if DEBUG
                    stopw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} SNPWriter; {1} SNPs writen in TabDelimited style output format; Time used: {2} ms", DateTime.Now.TimeOfDay, snps.Count, stopw.ElapsedMilliseconds));
#endif
                }
            }



            /// <summary>
            /// Reads SSRs from a TabDelimitedSSR-File
            /// </summary>
            public class SNPReader
            {
                private StreamReader sr;
                private readonly int bufferSize;
                private bool? snpAlleleMode = null;




                public SNPReader(StreamReader sr, int bufferSize)
                {
                    this.sr = sr;
                    this.bufferSize = bufferSize;

                }
                public SNPReader(StreamReader sr)
                {
                    this.sr = sr;
                    this.bufferSize = 5000;

                }

                public List<SNP> ReadSNPs()
                {
                    List<SNP> snpList = new List<SNP>();
                    string line = "";



#if DEBUG
                    System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                    stopw.Start();
#endif
                    //As long as there are lines in rhe file proceed!
                    while ((snpList.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                    {
                        //Get rid of space
                        line = line.Trim();
                        if (line == "" || line.StartsWith("#")) continue;
                        else if(snpAlleleMode==null && line.StartsWith(";"))
                        {
                            //sw.WriteLine(String.Format(";Mode {0}",snps[0].IsSNPAllele==true?"SNP-Alleles":"SNPs"));
                            if(line.StartsWith(";Mode"))
                            {
                                string test = line.Split(new char[] { ' ' })[1];
                                if (test == "halfSNP") this.snpAlleleMode = false; //(String.Format(";Mode {0}",snps[0].IsFullSNP==true?"fullSNP":"halfSNP"));
                                else if (test == "fullSNP") this.snpAlleleMode = true;
                                else throw new InvalidDataException("Invalid SNP mode");
                            }
                        }
                        else
                        {
                            snpList.Add(ParseTabDelimitedSNP(line));

                        }

                    }
#if DEBUG
                    stopw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} SNPReader; {1} SNPs read in {2} ms", DateTime.Now.TimeOfDay, snpList.Count, stopw.ElapsedMilliseconds));
#endif

                    return snpList;
                }



                //Extracts the SSR related information from the sputnikInfo string and stores the SSR in the database
                private SNP ParseTabDelimitedSNP(string toParse)
                {
                    //      0            1           2          3          4                    5               6        7                    8                      9                      10                      11                               12
                    //database_name\tquery_name\tstrand\tdatabase_char\tquery_char\tposition_database\tposition_query\tvalid\tdistance_from_alignment_ends\talignment_Quality_Token\tsequence_quality_at_site\tsequence_quality_in_neighborhood\tallele_id");
             
                    string[] division = toParse.Split(new char[] { '\t' });
                    if (division.Length <7) throw new Exception("Error during reading of Tab Delimited SNP-File.");
                    

                    SNP s = new SNP(division[0]
                        , division[1]
                        , division[3][0]
                        , division[4][0]
                        , division[5]
                    , division[6]);

                    s.PlusPlusStrand = division[2] == "+/+" ? true : false;
                    s.IsFullSNP = this.snpAlleleMode.Value;
                    s.Valid = division[7] == "True" ? true : false;
                    

                    if (division.Length > 8)
                    {
                        if(division[8]!=null && division[8]!="")  s.DistanceFromAlignmentEnd = Convert.ToInt32(division[8]);
                    }

                    if (division.Length > 9)
                    {
                        if (division[9] != null && division[9] != "") s.AlignmentQualityTokens = Convert.ToInt32(division[9]);
                    }
                    if (division.Length > 10)
                    {
                        if (division[10] != null && division[10] != "") s.SequenceQualityAtSite = Convert.ToInt32(division[10]);
                    }
                    if (division.Length > 11)
                    {
                        if (division[11] != null && division[11] != "") s.SequenceQualityInNeighborhood = Convert.ToInt32(division[11]);
                    }
                    if(division.Length>12)
                   {
                       division[12] = division[12].Trim();
                       if (division[12] != null && division[12] != "") if (division[12] != "") s.AlleleID = division[12];
                   }


                    


                    //Add the Microsatellite
                    return s;

                }

            }


            public class SNPSiteIO
            {
                /// <summary>
                /// Write the Loci into a file, info for the subset of SNPs at the locus
                /// </summary>
                /// <param name="sw"></param>
                /// <param name="snpSites"></param>
                public static void WriteSNPSites(StreamWriter sw, List<SNPSite> snpSites)
                {
                    sw.WriteLine("#SNP LOCI");
                    sw.WriteLine("#database_parent\tdatabase_position\tcount_subset\tcount_alleles_subset\tcount_pic_subset");
                    sw.WriteLine("#only loci which having a subset count larger than zero are writen");
                    sw.WriteLine();

                    foreach (SNPSite s in snpSites)
                    {
                        if (s.Count_Valid > 0) //Write the SNP site only if the counts are larger than zero
                        {
                            sw.WriteLine(String.Format("{0}\t{1}\t{2}\t{3}\t{4:f}"
                                , s.DatabaseParent
                                , s.SiteID_Database
                                , s.Count_Valid
                                , s.Count_Alleles
                                ,s.PIC));
                        }
                    }
                }


                public static List<SNPSite> ReadSNPSites(StreamReader sr)
                {
                    List<SNPSite> snpSites = new List<SNPSite>();
                    string line = "";

                    //As long as there are lines in rhe file proceed!
                    while ((line = sr.ReadLine()) != null)
                    {
                        //Get rid of space
                        line = line.Trim();
                        if (line == "" || line.StartsWith("#")) continue;
                        else
                        {
                            //              0                   1               2           3                           4                   5           6               7           8
                            // ("#database_parent\tdatabase_position\tcount_subset\tcount_alleles_subset\tcount_pic_subset");
                            string[] split = line.Split(new char[] { '\t' });
                            SNPSite s = new SNPSite(split[0],split[1]);
                            snpSites.Add(s);
                        }
                    }

                    return snpSites;

                }
            }


        }

        namespace Repetitive
        {
            using Bio.Repetitive;

            /// <summary>
            /// Static class providing basic functionallity for IO of the RepetitiveSequenceIndex
            /// </summary>
            public static class LogRepetitiveIO
            {

                /// <summary>
                /// Write repetititve sequence index into a file
                /// </summary>
                /// <param name="repS"></param>
                /// <param name="path"></param>
                /// <returns></returns>
                public static bool WriteRepetitiveSequenceIndexUshort(RepetitiveSequenceIndexUshort repS, string path)
                {
                    FileStream fs = new FileStream(path, FileMode.Create, FileAccess.Write);

                    DNADictionary<ushort> dict = repS.Dictionary;
                    int wordsize = repS.Wordsize;
                    ///The first position is the word size!!!
                    fs.Write(new byte[] { (byte)wordsize }, 0, 1);


                    //All other positions are the index counts!!!
                    for (int i = 0; i < dict.Count; i++)
                    {
                        //fs.Position=i*4; should not be necessary
                        fs.Write(BitConverter.GetBytes(dict.GetValueForHash(i)), 0, 2);
                    }
                    fs.Flush();
                    fs.Close();

                    return true;

                }

                /// <summary>
                /// Read repetitive sequence index from a file
                /// </summary>
                /// <param name="path"></param>
                /// <returns></returns>
                public static RepetitiveSequenceIndexUshort ReadRepetitiveSequenceIndexUshort(string path)
                {
                    FileStream fs = new FileStream(path, FileMode.Open, FileAccess.Read);

                    byte[] b = new byte[1];
                    fs.Read(b, 0, 1);
                    int wordsize = (int)b[0];
                    DNADictionary<ushort> dict = new DNADictionary<ushort>(wordsize);
                    b = new byte[2];

                    int dictsize = (int)Math.Pow(4.0, wordsize);
                    for (int i = 0; i < dictsize; i++)
                    {
                        fs.Read(b, 0, 2);
                        dict.Add(i, BitConverter.ToUInt16(b, 0));
                    }
                    fs.Close();

                    return new RepetitiveSequenceIndexUshort(dict, wordsize);
                }



                /// <summary>
                /// Write repetititve sequence index into a file
                /// </summary>
                /// <param name="repS"></param>
                /// <param name="path"></param>
                /// <returns></returns>
                public static bool WriteRepetitiveSequenceIndexInt(RepetitiveSequenceIndexInt repS, string path)
                {
                    FileStream fs = new FileStream(path, FileMode.Create, FileAccess.Write);

                    DNADictionary<int> dict = repS.Dictionary;
                    int wordsize = repS.Wordsize;
                    ///The first position is the word size!!!
                    fs.Write(new byte[] { (byte)wordsize }, 0, 1);


                    //All other positions are the index counts!!!
                    for (int i = 0; i < dict.Count; i++)
                    {
                        //fs.Position=i*4; should not be necessary
                        fs.Write(BitConverter.GetBytes(dict.GetValueForHash(i)), 0, 4);
                    }
                    fs.Flush();
                    fs.Close();

                    return true;

                }

                /// <summary>
                /// Read repetitive sequence index from a file
                /// </summary>
                /// <param name="path"></param>
                /// <returns></returns>
                public static RepetitiveSequenceIndexInt ReadRepetitiveSequenceIndexInt(string path)
                {
                    FileStream fs = new FileStream(path, FileMode.Open, FileAccess.Read);

                    byte[] b = new byte[1];
                    fs.Read(b, 0, 1);
                    int wordsize = (int)b[0];
                    DNADictionary<int> dict = new DNADictionary<int>(wordsize);
                    b = new byte[4];

                    int dictsize = (int)Math.Pow(4.0, wordsize);
                    for (int i = 0; i < dictsize; i++)
                    {
                        fs.Read(b, 0, 4);
                        dict.Add(i, BitConverter.ToInt32(b, 0));
                    }
                    fs.Close();

                    return new RepetitiveSequenceIndexInt(dict, wordsize);
                }

                
            }


        }



        namespace Seq
        {
            public interface INucSeqInfoReader
            {
                List<INucSeqInfo> GetSequences();
                void Close();
            }
            public interface INucSeqInfoWriter
            {
              void WriteSequenceInfos(List<INucSeqInfo> sequenceInfos);
                void Close();
            }



            /// <summary>
            /// Writes a collection of NucSeqInfos into a file
            /// </summary>
            public class NucSeqInfoWriter : INucSeqInfoWriter
            {
                private StreamWriter sw;
                private bool introductionWriten = false;

                public  NucSeqInfoWriter(StreamWriter sw)
                {
                    this.sw = sw;
                }

                 public NucSeqInfoWriter(string path):this(new StreamWriter(path,false,Encoding.ASCII))
                {
                }

                public void WriteSequenceInfos(List<INucSeqInfo> sequenceInfos)
                {
                    if (!introductionWriten)
                    {
                        sw.WriteLine("%>Sequence names");
                        sw.WriteLine("%;Comment");
                        sw.WriteLine("%nfo:\tLength\tIsUpperCase\tParentName\tStart\tEnd\tCount_A\tCount_T\tCount_U\tCount_C\tCount_G\tCount_N\tCount_Special\tCount_Invalid\tIndividual_sequences");
                        sw.WriteLine();
                        introductionWriten = true;
                    }

                    foreach (INucSeqInfo ns in sequenceInfos)
                    {

                        for(int i=0; i<ns.Names.Count;i++)
                        {
                            sw.WriteLine(String.Format(">{0}",ns.Names[i]));
                        }

                        if(ns.Comment!=null && ns.Comment !="") sw.WriteLine(String.Format(";{0}",ns.Comment));

                        //                                                                                                                 0        1           2           3           4       5           6           7           8           9           10          11                  12                  13
                        sw.WriteLine(String.Format("nfo:\t{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}\t{12}\t{13}", ns.Length, ns.IsUpper, ns.ParentName, ns.Start, ns.End, ns.Count_A, ns.Count_T, ns.Count_U, ns.Count_C, ns.Count_G, ns.Count_N, ns.Count_Special, ns.Count_Invalid,ns.SequenceCount));
                        sw.WriteLine();

                    }
                }
                public void Close()
                {
                    this.sw.Close();
                }


            }

            public class NucSeqInfoReader : INucSeqInfoReader
            {
                StreamReader sr;
                private int bufferSize = 5000;
                public NucSeqInfoReader(StreamReader sr)
                {
                    this.sr = sr;
                }
                public NucSeqInfoReader(string path)
                    : this(new StreamReader(path, Encoding.ASCII))
                {
                }

                public NucSeqInfoReader(StreamReader sr, int bufferSize):this(sr)
                {
                    this.bufferSize = bufferSize;
                }

                public NucSeqInfoReader(string path, int bufferSize)
                    : this(new StreamReader(path, Encoding.ASCII), bufferSize)
                {
                }

                public List<INucSeqInfo> GetSequences()
                {
                    string line = "";
                    List<INucSeqInfo> temp = new List<INucSeqInfo>();
                    NucSeqInfo ns = new NucSeqInfo();

                    //bool stored =false;

                    while (temp.Count<bufferSize && (line = sr.ReadLine()) != null)
                    {
                        line = line.Trim();
                        if (line == "" || line.StartsWith("%")) continue;

                        if (line.StartsWith(">"))
                        {
                            if (ns.Names == null) ns.Name = line.Substring(1, line.Length - 1);
                            else ns.Names.Add(line.Substring(1,line.Length-1));
                        }
                        else if (line.StartsWith(";"))
                        {
                            ns.Comment= line.Substring(1,line.Length-1);
                        }
                        else if(line.StartsWith("nfo:"))
                        {
                            //0     1       2           3           4       5   6       7       8       9       10          11      12              13              14                      

                            //nfo:	Length	IsUpperCase	ParentName	Start	End	Count_A	Count_T	Count_U	Count_C	Count_G	Count_N	Count_Special	Count_Invalid	Individual_sequences

                            string[] splits = line.Split(new char[] { '\t' });
                            ns.Length = Convert.ToInt32(splits[1]);
                            ns.IsUpper = splits[2] == "True" ? true : false;
                            ns.ParentName = splits[3];

                            ns.Start = (splits[4] == "" || splits[4] == "null") ? (int?)null : Convert.ToInt32(splits[4]);
                            ns.End = (splits[5] == "" || splits[5]=="null")? (int?)null : Convert.ToInt32(splits[5]);
                            ns.Count_A = Convert.ToInt32(splits[6]);
                            ns.Count_T = Convert.ToInt32(splits[7]);
                            ns.Count_U = Convert.ToInt32(splits[8]);
                            ns.Count_C = Convert.ToInt32(splits[9]);
                            ns.Count_G = Convert.ToInt32(splits[10]);
                            ns.Count_N = Convert.ToInt32(splits[11]);
                            ns.Count_Special = Convert.ToInt32(splits[12]);
                            ns.Count_Invalid = Convert.ToInt32(splits[13]);
                            ns.SequenceCount = Convert.ToInt32(splits[14]);
                            temp.Add(ns);
                            ns = new NucSeqInfo();
                        }
                        else throw new ArgumentOutOfRangeException("Somethings is wrong with the nucseqinfo");
                    }



                    //if (!stored)return new List<INucSeqInfo>();
                  return temp;


                }

                public void Close()
                {
                    this.sr.Close();

                }

            }



            public class QualityFileReader 
            {
                StreamReader sr;
                private int bufferSize = 5000;
                private string name;
                ISequenceValidator validator=new SequenceValidator_AllValid();
                bool ignoreInvalid;
                ISequenceContainer seqTemplate = SequenceFactory.GetDefaultSequence();


                public QualityFileReader(StreamReader sr)
                {
                    this.sr = sr;
                }
                public QualityFileReader(StreamReader sr, ISequenceValidator validator)
                {
                    this.sr = sr;
                    this.validator = validator;
                }
                public QualityFileReader(string path)
                    : this(new StreamReader(path, Encoding.ASCII))
                {
                }

                public QualityFileReader(StreamReader sr, int bufferSize)
                {
                    this.sr = sr;
                    this.bufferSize = bufferSize;
                }

                public QualityFileReader(StreamReader sr, ISequenceValidator validator, int bufferSize)
                {
                    this.sr = sr;
                    this.bufferSize = bufferSize;
                    this.validator = validator;
                }


                public List<QualitySequence> GetSequences()
                {
                    /*
                    >ES6HDNH02EISLO
                    28 28 27 27 34 28 28 27 27 28 23 33 25 26 35 28 27 27 28 28 34 27 27 28 26 27 32 25 26 28 27 34 27 27 27 28 28 27 27 27 27 26 28 35 28 34 28 27 37 33 17 27 27 33 25 28 34 27 28 33 25 28 28 34 27 34 27 25 27 27 27 27 27 27 34 27 28 34 27 27 38 34 22 09 28 28 27 27 27 28 34 28 25 27 27 27 28 33 26 28 28 27 28 28 27 28 27 28 33 26 28 28 27 26 28 27 24 28 28 38 34 22 10 27 25 28 28 27 37 33 15 28 27 27 37 33 16 27 27 27 28 27 27 28 27 27 37 33 16 28 27 26 23 26 33 26 25 32 24 27 35 31 12 35 28 27 27 34 28 26 32 24 36 32 13 31 24 21 31 23 27 35 31 11 28 31 26 03 27 28 25 38 34 24 16 09 2 27 27 23 24 24 28 37 33 20 06 21 20 35 30 11 27 28 37 33 16 27 37 33 16 26 33 26 36 32 13 37 33 16 20 27 27 21 37 33 17 20 23 37 32 14 37 33 15 27 27 18 27 27 37 33 16 18 26 34 28 37 33 15 25 20 28 26 18 25 32 24 21
                    */
                    string line = "";
                    List<QualitySequence> sequences = new List<QualitySequence>();
                    ISequenceContainer sequence = seqTemplate.NewSequence();
#if DEBUG
                    System.Diagnostics.Stopwatch sw = new System.Diagnostics.Stopwatch();
                    sw.Start();
#endif

                    while (sequences.Count < bufferSize && (line = sr.ReadLine()) != null)
                    {
                        line = line.Trim();
                        if (line == "") continue;

                        if (line.StartsWith(">"))
                        {

                            if (name != "" && sequence.Length != 0)
                            {
                                //Adjust bufferSize for large sequences
                                if (sequence.Length > 2000000 && bufferSize > 1) bufferSize = 1;
                                else if (sequence.Length > 200000 && bufferSize > 10) bufferSize = 10;

                                if (validator.IsValid(sequence)) sequences.Add(new QualitySequence(name, sequence));
                                else if (!ignoreInvalid) throw new InvalidSequenceException(validator);
                            }

                            //Create new name and sequence
                            name = line.Substring(1, line.Length - 1);
                            sequence = seqTemplate.NewSequence();
                        }
                        else
                        {
                            string[] division=line.Split(new char[]{' '});
                            for (int i = 0; i < division.Length; i++)
                            {
                                if(division[i]!="") sequence.Append((byte)Convert.ToInt32(division[i]));
                            }

                        }
                    }
                    //Also store last file
                    if (name != "" && sequence.Length != 0)
                    {
                        if (validator.IsValid(sequence)) sequences.Add(new QualitySequence(name, sequence));
                        else if (!ignoreInvalid) throw new InvalidSequenceException(validator);
                    }

#if DEBUG
                    sw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} QualityFileReader; {1} files loaded in {2} ms; First file contained {3} nucleotides; Sequence used: {4}", DateTime.Now.TimeOfDay, sequences.Count, sw.ElapsedMilliseconds, sequences.Count <= 0 ? 0 : sequences[0].Length, seqTemplate.GetType().ToString()));
#endif

                    //if (!stored)return new List<INucSeqInfo>();
                    return sequences;
                }

                /// <summary>
                /// Get or set a sequence prototype
                /// </summary>
                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return this.seqTemplate ;
                    }
                    set
                    {
                        this.seqTemplate = value;
                    }
                }

                public void Close()
                {
                    this.sr.Close();
                }

            }

            /// <summary>
            /// Write QualitySequences into a file
            /// </summary>
            public class QualityFileWriter
            {
                private StreamWriter sw;
                int leng = 50;
                public QualityFileWriter(string path) : this(new StreamWriter(path, false, Encoding.ASCII)) { }
                public QualityFileWriter(StreamWriter sw)
                {
                    this.sw = sw;
                }

                public void WriteSequences(List<QualitySequence> sequences)
                {
#if DEBUG
                    System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                    stopw.Start();
#endif
                    foreach (QualitySequence qs in sequences)
                    {
                        int count = 0;
                        // string line = "";
                        ISequenceContainer seq = qs.Sequence;
                        sw.WriteLine('>' + qs.Name);

                        while (seq.Length >= count + leng)
                        {
                            //seq.ToString(count, leng)
                            StringBuilder sb = new StringBuilder("");
                            for (int i = 0; i < leng; i++)
                            {
                                sb.Append((int)seq[count + i]);
                                if (i < leng - 1) sb.Append(' ');
                            }
                            sw.WriteLine(sb.ToString());
                            count += leng;

                        }
                        StringBuilder sb2 = new StringBuilder();
                        for(int i=0;i<seq.Length-count;i++)
                        {
                             sb2.Append((int)seq[count + i]);
                             if (i < seq.Length - count - 1) sb2.Append(' ');
                        }
                        sw.WriteLine(sb2.ToString());
                        sw.WriteLine();
                    }

#if DEBUG
                    stopw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} QualityFileWriter; {1} files saved in {2} ms; First file contained {3} quality scores; Sequence used: {4}", DateTime.Now.TimeOfDay, sequences.Count, stopw.ElapsedMilliseconds,
                        sequences.Count <= 0 ? 0 : sequences[0].Length, sequences.Count <= 0 ? "no sequence" : sequences[0].Sequence.GetType().ToString()));


#endif
                }

                public void Close()
                {
                    this.sw.Close();
                }


            }


            public interface ISequenceWriter<TSimpleSequence> where TSimpleSequence: ISimpleSequence
            {
                void WriteSequences(List<TSimpleSequence> sequences);
                void Close();
            }


            /// <summary>
            /// Has to be implemented by SimpleSequence derived sequences.
            /// Implementation of this interface allows them to be read with ISequenceReader 
            /// </summary>
            /// <typeparam name="TSimpleSequence"></typeparam>
            public interface IFastaFileReadable<TSimpleSequence> where TSimpleSequence : ISimpleSequence
            {
                TSimpleSequence NewInstance(string name, string comment, ISequenceContainer sequence);
                TSimpleSequence NewInstance(string name, string parentName, ISequenceContainer sequence, int startPosition, int endPosition);


            }

            public interface ISequenceReader<TSimpleSequence> where TSimpleSequence: ISimpleSequence
            {
                List<TSimpleSequence> GetSequences();
                ISequenceContainer SequencePrototype { get;set;}
                void Close();
            }


            /// <summary>
            /// Writes a simple nucleotide Sequence in a fasta file, heading the recommendation that the one line should have a length of 
            /// at most 50 nucleotides
            /// </summary>
            public class FastaWriter<TSimpleSequence> : ISequenceWriter<TSimpleSequence> where TSimpleSequence: ISimpleSequence
            {
                private StreamWriter sw;
                public FastaWriter(string path): this(new StreamWriter(path,false,Encoding.ASCII)){}
                public FastaWriter(StreamWriter sw)
                {
                    this.sw = sw;
                }

                /// <summary>
                /// Writes nucleotide sequences into a fasta-style output file.
                /// Since the fasta definition also contains a comment line (starts with ;) the comments are also writen to the output file
                /// If the .Comment property is null or "" no comment line will be written
                /// </summary>
                /// <param name="sequences"></param>
                public void WriteSequences(List<TSimpleSequence> sequences)
                {
#if DEBUG
                    System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                    stopw.Start();
#endif
                    foreach (TSimpleSequence ns in sequences)
                    {
                        int count = 0;
                        // string line = "";
                        ISequenceContainer seq = ns.Sequence;
                        sw.WriteLine('>' + ns.Name);
                        if(ns.Comment!=null && ns.Comment!="") sw.WriteLine(';' + ns.Comment);

                        while (seq.Length >= count + 50)
                        {

                            sw.WriteLine(seq.ToString(count, 50));

                            count += 50;

                        }

                        sw.WriteLine(seq.ToString(count, seq.Length - count));
                        sw.WriteLine();
                    }

#if DEBUG
                    stopw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} FastaWriter; {1} files saved in {2} ms; First file contained {3} nucleotides; Sequence used: {4}", DateTime.Now.TimeOfDay, sequences.Count, stopw.ElapsedMilliseconds,
                        sequences.Count <= 0 ? 0 : sequences[0].Length, sequences.Count <= 0 ? "no sequence" : sequences[0].Sequence.GetType().ToString()));


#endif
                }

                public void Close()
                {
                    this.sw.Close();
                }


            }


            /// <summary>
            /// Writes a simple nucleotide Sequence into a byte-file;
            /// Attention the byte(255) and byte(254) are used for internal puposes and mark the end of the sequence and the end of the sequence-name respectively
            /// the first 4000 bytes are reserved for the name of the sequence
            /// the second 4000 bytes are reserved for comments
            /// </summary>
            public class ByteWriter<TSimpleSequence> : ISequenceWriter<TSimpleSequence> where TSimpleSequence : ISimpleSequence
            {
                private FileStream fs;
                public ByteWriter(string path) : this(new FileStream(path,FileMode.Create,FileAccess.Write)) { }
                public ByteWriter(FileStream fs)
                {
                    this.fs = fs;
                }

                /// <summary>
                /// Writes nucleotide sequences into a byte-output file.
                /// byte(255) is reserved and marks the end of a sequence 
                /// first 4000 bytes are for the name the next 4000 bytes for the comment
                /// </summary>
                /// <param name="sequences"></param>
                public void WriteSequences(List<TSimpleSequence> sequences)
                {
#if DEBUG
                    System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                    stopw.Start();
#endif
                    foreach (TSimpleSequence ns in sequences)
                    {
                        
                        string seQName=ns.Name;
                        string seqComment=ns.Comment;

                        byte[] name = new byte[seQName.Length + 1];
                        byte[] comment = new byte[1];
                        if(seqComment!=null)  comment = new byte[seqComment.Length+1];

                        //First write name of the sequence
                        for (int i = 0; i < seQName.Length; i++)
                        {
#if DEBUG
                            if ((((byte)seQName[i]) == (byte)255) || (((byte)seQName[i]) == (byte)254)) throw new Exception("crap");
#endif
                          name[i] = (byte)seQName[i];
                            
                        }
                        name[name.Length-1]=(byte)254;

                        //Second write the comment
                        for(int i=0; i<(comment.Length-1);i++)
                        {
#if DEBUG
                            if ((((byte)seqComment[i]) == (byte)255) || (((byte)seqComment[i]) == (byte)254)) throw new Exception("crap");
#endif
                                comment[i] = (byte)seqComment[i];
                        }
                        comment[comment.Length - 1] = (byte)254;

                        fs.Write(name, 0, name.Length);
                        fs.Write(comment, 0, comment.Length);

                        //Third write the sequence to the file
                        ISequenceContainer seq = ns.Sequence;
                        for (int i = 0; i < seq.Length; i++)
                        {
#if DEBUG
                            if ((((byte)seq[i]) == (byte)255) || (((byte)seq[i]) == (byte)254)) throw new Exception("crap");
#endif

                            fs.WriteByte((byte)seq[i]);
                        }
                       fs.Write(new byte[]{(byte)255},0,1);
                    }

#if DEBUG
                    stopw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} ByteWriter; {1} files saved in {2} ms; First file contained {3} nucleotides; Sequence used: {4}", DateTime.Now.TimeOfDay, sequences.Count, stopw.ElapsedMilliseconds,
                        sequences.Count <= 0 ? 0 : sequences[0].Length, sequences.Count <= 0 ? "no sequence" : sequences[0].Sequence.GetType().ToString()));


#endif
                }

                public void Close()
                {
                    this.fs.Close();
                }
            }



            /// <summary>
            /// Creates a number of random sequences, can be plugged into each application instead of an actual fasta reader
            /// </summary>
            public class RandomGenerater_NucleotideSequence : ISequenceReader<NucleotideSequence>
            {
                private int contentGC;
                private int count;
                private int actualCount = 0;
                private int length;
                private int bufferSize = 5000;
                private ISequenceContainer sequencePrototype = SequenceFactory.GetDefaultSequence();

                /// <summary>
                /// Generate a number of sequences with a given CG-content and length, randomly
                /// </summary>
                /// <param name="count">number of sequences to generate</param>
                /// <param name="length">length of the sequences to generate</param>
                /// <param name="contentGC">the GC content of the sequences to generate, number between 1 - 100</param>
                /// <param name="buffer">buffer size, how many sequences should be generated at one run; if the sequence length exceeds 200.000bp this feature will be automatically set, to either 10 or 1</param>
                public RandomGenerater_NucleotideSequence(int count, int length, int contentGC, int buffer):this(count,length,contentGC)
                {
                    if(this.length<200000) this.bufferSize = buffer;

                }


                /// <summary>
                /// Generate a number of sequences with a given CG-content and length, randomly
                /// </summary>
                /// <param name="count">number of sequences to generate</param>
                /// <param name="length">length of the sequences to generate</param>
                /// <param name="contentGC">the GC content of the sequences to generate, number between 1 - 100</param>
                public RandomGenerater_NucleotideSequence(int count, int length, int contentGC)
                {
                    this.contentGC = contentGC;
                    this.length = length;
                    this.count = count;

                    if (length > 2000000) this.bufferSize = 1;
                    else if (length > 200000) this.bufferSize = 10;
                }

                /// <summary>
                /// Returns a list of randomly generarted nucleotide sequences
                /// </summary>
                /// <returns></returns>
                public List<NucleotideSequence> GetSequences()
                {
                    List<NucleotideSequence> ret = new List<NucleotideSequence>(bufferSize);
#if DEBUG
                    System.Diagnostics.Stopwatch sw = new System.Diagnostics.Stopwatch();
                    sw.Start();
#endif

                    //as long as new sequences are required -> produce them
                    while (ret.Count < bufferSize && actualCount < count)
                    {
                        ISequenceContainer newSeq = SequenceFactory.GetDefaultSequence();

                        Random rand = new Random((int)DateTime.Now.Ticks);
                        //Formula
                        // gcTotal * lengthTotal = gcDone * lengthDone + gcToSet * lengthToSet
                        // => gcToSet = (gcTotal * lengtTotal - gcActual *lengthActual)/lengthToSet
                        double gcActual = 0.0;
                        int gcCount=0;
                        double gcTotal = (double)contentGC;
                        int gcToSet=(int)gcTotal;


                        for (int i = 0; i < length; i++)
                        {

#if DEBUG
                            int rand1 = rand.Next(1, 101);
                            int rand2 = rand.Next(1, 3);
                            if (rand1 < 1 || rand1 > 100) throw new Exception("fuck random generater");
                            if (rand2 < 1 || rand2 > 2) throw new Exception("they suck");
#endif
                            //lower boundary is inclusive, upper is exclusive
                            if(rand.Next(1,101)<=gcToSet)
                            {
                                //New G or C
                                if (rand.Next(1, 3) == 1)
                                {
                                    newSeq.Append('C');
                                }
                                else
                                {
                                    newSeq.Append('G');
                                }

                                //Updated gcActual
                                gcCount++;
                                gcActual = ((double)gcCount)*100.0 / (double)newSeq.Length;

                            }
                            else
                            {
                                //new A or T
                                if (rand.Next(1, 3) == 1)
                                {
                                    newSeq.Append('A');
                                }
                                else
                                {
                                    newSeq.Append('T');
                                }

                                //update gcActual
                                gcActual = ((double)gcCount) * 100.0 / (double)newSeq.Length;

                            }

                              gcToSet  = (int)(((gcTotal*(double)length)-(gcActual*(double)newSeq.Length))/((double)(length-newSeq.Length)));
                        }

                        ret.Add(new NucleotideSequence(String.Format("random_sequence;gc-content_{0};sequence_number_{1}",contentGC,actualCount), newSeq));
                        actualCount++;
                    }

#if DEBUG
                    sw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} RandomGenerator_NucleotideSequence; {1} sequenceces generated in {2} ms; First file contained {3} nucleotides; Sequence used: {4}", DateTime.Now.TimeOfDay, ret.Count, sw.ElapsedMilliseconds, ret.Count <= 0 ? 0 : ret[0].Length, this.sequencePrototype.GetType().ToString()));

#endif


                    return ret;

                }

                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return sequencePrototype;
                    }
                    set
                    {
                        this.sequencePrototype = value;
                    }
                }

                public void Close()
                {
                   
                }


            
            }


            /// <summary>
            /// Reads nucleotide sequences, and chops the sequence into chunks of the given size.
            /// A overhead might be specified (i.e. a small nucleotide sequence that overlaps with the previous sequence).
            /// All chunks derived from a large sequence have the same fasta-ID and comment.
            /// </summary>
            public class FastaReaderSequenceChopper<TSimpleSequence> : ISequenceReader<TSimpleSequence> 
                where TSimpleSequence:ISimpleSequence,IFastaFileReadable<TSimpleSequence>,new()
            {
                private readonly ISequenceValidator validator;
                private readonly bool ignoreInvalid;

                private StreamReader sr;
                private ISequenceContainer seqTemplate;
                private const int bufferSize = 1;
                private List<TSimpleSequence> sequences;

                //Variables necessary when operating
                private string name;
                private string comment;
                private ISequenceContainer sequence;
                private string previousOverhead = "";

                //Special FastaReader_Chopper variables
                private readonly int sequenceSize;
                private readonly int overhead;
                private TSimpleSequence templateSimpleSequence= new TSimpleSequence();


                public FastaReaderSequenceChopper(string path, ISequenceValidator validator, bool ignoreInvalid, int sequenceSize, int overhead):
                    this(new StreamReader(path,Encoding.ASCII),validator,ignoreInvalid,sequenceSize,overhead){}
                public FastaReaderSequenceChopper(StreamReader sr, ISequenceValidator validator, bool ignoreInvalid, int sequenceSize, int overhead):this()
                {
                    this.sr = sr;
                    this.ignoreInvalid = ignoreInvalid;
                    this.validator = validator;
                    sequences = new List<TSimpleSequence>(bufferSize);

                    //setting operating varibles to their default values
                    name = "";

                    //Setting special variables
                    if (overhead >= sequenceSize) throw new ArgumentOutOfRangeException("Overhead > Sequence size");
                    this.sequenceSize = sequenceSize;
                    this.overhead = overhead;
                }

                private FastaReaderSequenceChopper()
                {
                    this.seqTemplate = SequenceFactory.GetDefaultSequence();
                    this.sequence = seqTemplate.NewSequence();
                }

                public FastaReaderSequenceChopper(string path, ISequenceValidator validator, bool ignoreInvalid):
                    this(new StreamReader(path,Encoding.ASCII), validator, ignoreInvalid){}
                public FastaReaderSequenceChopper(StreamReader sr, ISequenceValidator validator, bool ignoreInvalid)
                    : this()
                {
                    this.sr = sr;
                    this.ignoreInvalid = ignoreInvalid;
                    this.validator = validator;
                    sequences = new List<TSimpleSequence>(bufferSize);

                    //setting operating varibles to their default values
                    name = "";

                    //
                    this.sequenceSize = 50000000;
                    this.overhead = 50;

                }

                public List<TSimpleSequence> GetSequences()
                {
                    sequences.Clear();
#if DEBUG
                    System.Diagnostics.Stopwatch sw = new System.Diagnostics.Stopwatch();
                    sw.Start();
#endif
                    string line = "";


                    sequence.Append(previousOverhead.Trim());
                    //As long as there are lines in rhe file proceed!
                    while (sequences.Count < bufferSize &&
                        ((line = sr.ReadLine()) != null))
                    {
                        //Get rid of space
                        line = line.Trim();
                        //If line is empty or starts with the comment symbol skip it
                        if (line == "" ) continue;

                        //The > resets
                        if (line.StartsWith(">"))
                        {
                            //Store old FastaFile
                            if (name != "" && !sequence.Equals(seqTemplate.NewSequence()))
                            {

                                if (validator.IsValid(sequence)) sequences.Add(templateSimpleSequence.NewInstance(name,comment, sequence));
                                else if (!ignoreInvalid) throw new InvalidSequenceException(validator);
                            }

                            //Create new name
                            sequence = seqTemplate.NewSequence();
                            previousOverhead = "";
                            name = line.Substring(1, line.Length - 1);
                            comment = null;

                        }
                        else if (line.StartsWith(";"))
                        {
                            comment = comment ?? ""; //Comment is the content of comment except comment=null then comment=""
                            comment += line.Substring(1, line.Length - 1);
                        }
                        else if (sequence.Length >= this.sequenceSize)//If large files
                        {
                            //at first append the actual line
                            sequence.Append(line);

                            //get overhead for the next file eg sequence.Length=200052; sequenceSize=200000; overhead=50
                            //200052+50-200000=102; start position = 199999-50=199949
                            //10-5=5 
                            //01234      0123456789
                            //TTTTTAAAAA AAAAATTTTT 20+5-10=15
                            previousOverhead = sequence.ToString(sequenceSize - overhead, sequence.Length + overhead - sequenceSize);


                            if (validator.IsValid(sequence)) sequences.Add(templateSimpleSequence.NewInstance(name,comment, sequence.SubSequence(0, sequenceSize)));
                            else if (!ignoreInvalid) throw new InvalidSequenceException(validator);

                            sequence = seqTemplate.NewSequence();


                        }
                        else sequence.Append(line);

                    }
                    //Also store last file
                    if (name != "" && !sequence.Equals(seqTemplate.NewSequence()))//ReadLine null
                    {

                        if (validator.IsValid(sequence)) sequences.Add(templateSimpleSequence.NewInstance(name, comment, sequence));
                        else if (!ignoreInvalid) throw new InvalidSequenceException(validator);

                        previousOverhead = "";
                        sequence = seqTemplate.NewSequence();
                    }

#if DEBUG
                    sw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} FastaReaderChopper_NucleotideSequence; {1} files loaded in {2} ms; First file contained {3} nucleotides; Specified sequnce size: {4}; Specified overhead: {5}; Sequence used: {6}",
                        DateTime.Now.TimeOfDay, sequences.Count, sw.ElapsedMilliseconds, sequences.Count <= 0 ? 0 : sequences[0].Length, this.sequenceSize, overhead, seqTemplate.GetType().ToString()));

#endif

                    return sequences;
                }



                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return this.seqTemplate;
                    }
                    set
                    {
                        this.seqTemplate = value;
                        this.sequence = seqTemplate.NewSequence();
                    }
                }

                public void Close()
                {
                    this.sr.Close();
                }

            }


            /// <summary>
            /// Reading fasta files, return a collection of INucSeqInfo;
            /// This class only reads general informations for nucleotide sequences contained in a fasta file like: are the characters uppercase; how many A,T,C,G etc
            /// Comments are also read (;) 
            /// The detail level may be specified; fullDetails = false -> only the fastaID and the comments are returned
            /// A ISequenceValidatior may be passed
            /// </summary>
            public class FastaReader_NucSeqInfo : INucSeqInfoReader
            {
                private StreamReader sr;
                private readonly ISequenceValidator validator;
                private readonly bool ignoreInvalid;
                private int bufferSize=5000;
                INucSeqInfo nsi = new NucSeqInfo("");
                private List<INucSeqInfo> sequences;
                private bool fullDetails = true;


                /// <summary>
                /// Constructor for reading only the fasta ID, the length and the fasta comment,
                /// all fasta sequences are valid
                /// </summary>
                public FastaReader_NucSeqInfo(string path)
                    : this(path, new SequenceValidator_AllValid(), false)
                {
                    this.fullDetails = false;
                }

                /// <summary>
                /// Constructor for reading only the fasta ID, the length and the fasta comment,
                /// all fasta sequences are valid
                /// </summary>
                public FastaReader_NucSeqInfo(StreamReader sr)
                    : this(sr, new SequenceValidator_AllValid(), false)
                {
                    this.fullDetails = false;
                }




                /// <summary>
                /// Constructor for reading only the fasta ID, the length and the fasta comment,
                /// all fasta sequences are valid
                /// </summary>
                public FastaReader_NucSeqInfo(string path, int bufferSize)
                    : this(path, new SequenceValidator_AllValid(), false,bufferSize)
                {
                    this.fullDetails = false;
                }

                /// <summary>
                /// Constructor for reading only the fasta ID, the length and the fasta comment,
                /// all fasta sequences are valid
                /// </summary>
                public FastaReader_NucSeqInfo(StreamReader sr,int bufferSize)
                    : this(sr, new SequenceValidator_AllValid(), false,bufferSize)
                {
                    this.fullDetails = false;
                }



                public FastaReader_NucSeqInfo(string path, ISequenceValidator validator, bool ignoreInvalid, int bufferSize)
                    : this(new StreamReader(path, Encoding.ASCII), validator, ignoreInvalid, bufferSize) { }
                public FastaReader_NucSeqInfo(StreamReader sr, ISequenceValidator validator, bool ignoreInvalid, int bufferSize)
                    : this()
                {
                    this.sr = sr;
                    this.validator = validator;
                    this.bufferSize = bufferSize;
                    this.ignoreInvalid = ignoreInvalid;
                    sequences = new List<INucSeqInfo>(bufferSize);
                }
                private FastaReader_NucSeqInfo()
                {

                }
                public FastaReader_NucSeqInfo(string path, ISequenceValidator validator, bool ignoreInvalid):this(new StreamReader(path,Encoding.ASCII), validator,ignoreInvalid){}
                public FastaReader_NucSeqInfo(StreamReader sr, ISequenceValidator validator, bool ignoreInvalid)
                    : this()
                {
                    this.sr = sr;
                    this.validator = validator;
                    this.ignoreInvalid = ignoreInvalid;
                    this.bufferSize = 5000;
                    sequences = new List<INucSeqInfo>(this.bufferSize);
                }



                public List<INucSeqInfo> GetSequences()
                {
                    sequences.Clear();
#if DEBUG
                    System.Diagnostics.Stopwatch sw = new System.Diagnostics.Stopwatch();
                    sw.Start();
#endif
                    string line = "";
                   // string name = "";
                    string comment = null;


                    //As long as there are lines in rhe file proceed!
                    while (sequences.Count < bufferSize &&
                        ((line = sr.ReadLine()) != null))
                    {
                        //Get rid of space
                        line = line.Trim();

                        //If line is empty or starts with the comment symbol skip the line
                        if (line == "") continue;

                        //The > resets
                        if (line.StartsWith(">"))
                        {
                            //Store old FastaFile
                            if (nsi.Name != "" && nsi.Length != 0) //
                            {
                                nsi.Comment = comment;
                                if (validator.IsValid(nsi)) sequences.Add(nsi.Copy());
                                else if (!ignoreInvalid) throw new InvalidSequenceException(validator);


                            }
                            nsi = new NucSeqInfo(line.Substring(1, line.Length - 1));
                            nsi.IsUpper = true;
                            comment = "";

                        }
                        else if (line.StartsWith(";"))
                        {
                            comment = comment ?? ""; //Comment is the content of comment except comment=null then comment=""
                            comment += line.Substring(1, line.Length - 1);
                        }
                        else
                        {
                            //In any case increase count the length
                            nsi.Length += line.Length;


                            if (fullDetails)
                            {
                                for (int i = 0; i < line.Length; i++)
                                {
                                    char c = line[i];
                                    if (Char.IsLower(c)) nsi.IsUpper = false;

                                    switch (c)
                                    {


                                        case 'a':
                                        case 'A': nsi.Count_A++;
                                            break;
                                        case 't':
                                        case 'T': nsi.Count_T++;
                                            break;

                                        case 'c':
                                        case 'C': nsi.Count_C++;
                                            break;
                                        case 'g':
                                        case 'G': nsi.Count_G++;
                                            break;
                                        case 'u':
                                        case 'U': nsi.Count_U++;
                                            break;
                                        case 'n':
                                        case 'N': nsi.Count_N++;
                                            break;



                                        //Special symbols:

                                        case 'Y':   //Pyrmidin (C or T)
                                        case 'y':
                                        case 'r':   //Purin (G or A)
                                        case 'R':
                                        case 'w':   //Weich (A or T)
                                        case 'W':
                                        case 'S':   //Stark(C or G)
                                        case 's':
                                        case 'k':   //Ketogruppe (G or T)
                                        case 'K':
                                        case 'm':   //Aminogruppe (A or C)
                                        case 'M':
                                        case 'B':   //Nicht A (C,G,T)
                                        case 'b':
                                        case 'd':   //Nicht C (A,T,G)
                                        case 'D':
                                        case 'h':   //Nicht G (A,T,C)
                                        case 'H':
                                        case 'v':   //Nicht T (A,C,G)
                                        case 'V': nsi.Count_Special++;
                                            break;
                                        default: nsi.Count_Invalid++;
                                            break;



                                    }

                                }
                            }
                            else
                            {
                                nsi.IsUpper = null;
                            }
                            
                        }

                    }
                    //Also store last file
                    if (nsi.Name != "" && nsi.Length != 0)//
                    {
                        nsi.Comment = comment;
                        if (validator.IsValid(nsi)) sequences.Add(nsi.Copy());
                        else if (!ignoreInvalid) throw new InvalidSequenceException(validator);
                        nsi = new NucSeqInfo("");
                    }

#if DEBUG
                    sw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} FastaReader_NucSeqInfo; {1} files loaded in {2} ms; First file contained {3} nucleotides; Full detail level: {4}", DateTime.Now.TimeOfDay, sequences.Count, sw.ElapsedMilliseconds, sequences.Count <= 0 ? 0 : sequences[0].Length,fullDetails));

#endif


                    return sequences;
                }

                public void Close()
                {
                    this.sr.Close();
                }


            }

            /// <summary>
            /// Reading fasta files, return a collection of INucleotideSequences;
            /// Comments are also read (;)
            /// A ISequenceValidatior may be passed
            /// </summary>
            public class FastaReader_Sequence<TSimpleSequence> : ISequenceReader<TSimpleSequence> 
                where TSimpleSequence: ISimpleSequence,IFastaFileReadable<TSimpleSequence>,new()
            {
                private StreamReader sr;
                private readonly ISequenceValidator validator;
                private readonly bool ignoreInvalid;
                private int bufferSize=5000;
                private string name;
                private TSimpleSequence templateSimpleSequence = new TSimpleSequence();

                /// <summary>
                /// seqTemplate is used as template for sequences. Acts like a factory, the NewSequence() mehtod is called to obtain a new ISequence of the same type
                /// Therefore the user does not have to know the type of the Sequence
                /// </summary>
                private ISequenceContainer seqTemplate = SequenceFactory.GetDefaultSequence();

                private List<TSimpleSequence> sequences;


                public FastaReader_Sequence(string path, ISequenceValidator validator, bool ignoreInvalid, int bufferSize) : this(new StreamReader(path, Encoding.ASCII), validator, ignoreInvalid, bufferSize) { }
                public FastaReader_Sequence(StreamReader sr, ISequenceValidator validator, bool ignoreInvalid, int bufferSize)
                {
                    this.sr = sr;
                    this.validator = validator;
                    this.bufferSize = bufferSize;
                    this.ignoreInvalid = ignoreInvalid;
                    name = "";
                    sequences = new List<TSimpleSequence>(bufferSize);
                }




                public FastaReader_Sequence(string path, ISequenceValidator validator, bool ignoreInvalid):this(new StreamReader(path,Encoding.ASCII),validator,ignoreInvalid)
                {
                 }
                public FastaReader_Sequence(StreamReader sr, ISequenceValidator validator, bool ignoreInvalid)
                {
                    name = "";
                    this.sr = sr;
                    this.validator = validator;
                    this.ignoreInvalid = ignoreInvalid;
                    sequences = new List<TSimpleSequence>(this.bufferSize);
                }

                /// <summary>
                /// Returns a number of sequences as specified through the buffer size.
                /// If the size of a sequence is greater than >200000 than the buffer size is set to 10
                /// if the size of the sequence is greater than 2000000 the buffer size for this file is set to 1
                /// </summary>
                /// <returns></returns>
                public List<TSimpleSequence> GetSequences()
                {
#if DEBUG
                    System.Diagnostics.Stopwatch sw =new System.Diagnostics.Stopwatch();
                    sw.Start();
#endif
                    //The Mark updated
                    sequences.Clear();
                    string line = "";
                    string comment = null;
                    ISequenceContainer sequence = seqTemplate.NewSequence();

                    //As long as there are lines in rhe file proceed!
                    while ((sequences.Count < bufferSize) &&
                        ((line = sr.ReadLine()) != null))
                    {
                        //Get rid of space
                        line = line.Trim();

                        //If line is empty or starts with the comment symbol skip the line
                        if (line == "") continue;

                        //The > resets
                        if (line.StartsWith(">"))
                        {
                            //Store old FastaFile
                            if (name != "" && sequence.Length != 0)
                            {
                                //Adjust bufferSize for large sequences
                                if (sequence.Length > 2000000 && bufferSize > 1) bufferSize = 1;
                                else if (sequence.Length > 200000 && bufferSize > 10) bufferSize = 10;

                                if (validator.IsValid(sequence))
                                {
                                    sequence.TrimExcess();
                                    sequences.Add(templateSimpleSequence.NewInstance(name, comment, sequence));
                                }
                                else if (!ignoreInvalid) throw new InvalidSequenceException(validator);
                            }

                            //Create new name and sequence
                            name = line.Substring(1, line.Length - 1);
                            sequence = seqTemplate.NewSequence();
                            comment = null;
                        }
                        else if (line.StartsWith(";"))
                        {
                            comment = comment ?? ""; //Comment is the content of comment except comment=null then comment=""
                            comment += line.Substring(1, line.Length - 1);
                        }
                        else
                        {
                            sequence.Append(line);
                        }

                    }
                    //Also store last file
                    if (name != "" && sequence.Length != 0)
                    {
                        if (validator.IsValid(sequence))
                        {
                            sequence.TrimExcess();
                            sequences.Add(templateSimpleSequence.NewInstance(name, comment, sequence));
                        }
                        else if (!ignoreInvalid) throw new InvalidSequenceException(validator);
                    }

#if DEBUG
                    sw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} FastaReader_NucleotideSequence; {1} files loaded in {2} ms; First file contained {3} nucleotides; Sequence used: {4}", DateTime.Now.TimeOfDay, sequences.Count,sw.ElapsedMilliseconds, sequences.Count <= 0 ? 0 : sequences[0].Length, seqTemplate.GetType().ToString()));


#endif


                    return sequences;
                }

                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return this.seqTemplate;
                    }
                    set
                    {
                        this.seqTemplate = value;
                    }
                }

                public void Close()
                {
                    this.sr.Close();
                }

            }


            /// <summary>
            /// Reading byte files, return a collection of INucleotideSequences;
            /// byte(255) denotes the end of a sequence
            /// the first 4000 bytes are reserved for the name of the sequence
            /// the second 4000 bytes are reserved for comments
            /// A ISequenceValidatior may be passed
            /// </summary>
            public class ByteReader_Sequence<TSimpleSequence> : ISequenceReader<TSimpleSequence>
                where TSimpleSequence : ISimpleSequence, IFastaFileReadable<TSimpleSequence>, new()
            {
                private FileStream fs;
                private readonly ISequenceValidator validator;
                private readonly bool ignoreInvalid;
                private int bufferSize = 5000;
                private TSimpleSequence templateSimpleSequence = new TSimpleSequence();


                /// <summary>
                /// seqTemplate is used as template for sequences. Acts like a factory, the NewSequence() mehtod is called to obtain a new ISequence of the same type
                /// Therefore the user does not have to know the type of the Sequence
                /// </summary>
                private ISequenceContainer seqTemplate = SequenceFactory.GetDefaultSequence();

                private List<TSimpleSequence> sequences;


                public ByteReader_Sequence(string path, ISequenceValidator validator, bool ignoreInvalid, int bufferSize) : this(new FileStream(path,FileMode.Open,FileAccess.Read), validator, ignoreInvalid, bufferSize) { }
              
                public ByteReader_Sequence(FileStream fs, ISequenceValidator validator, bool ignoreInvalid, int bufferSize)
                {
                    this.fs = fs;
                    this.validator = validator;
                    this.bufferSize = bufferSize;
                    this.ignoreInvalid = ignoreInvalid;

                    sequences = new List<TSimpleSequence>(bufferSize);
                }




                public ByteReader_Sequence(string path, ISequenceValidator validator, bool ignoreInvalid)
                    : this(new FileStream(path, FileMode.Open, FileAccess.Read), validator, ignoreInvalid)
                {
                }
                public ByteReader_Sequence(FileStream fs, ISequenceValidator validator, bool ignoreInvalid)
                {
 
                    this.fs = fs;
                    this.validator = validator;
                    this.ignoreInvalid = ignoreInvalid;
                    sequences = new List<TSimpleSequence>(this.bufferSize);
                }

                /// <summary>
                /// Returns a number of sequences as specified through the buffer size.
                /// If the size of a sequence is greater than >200000 than the buffer size is set to 10
                /// if the size of the sequence is greater than 2000000 the buffer size for this file is set to 1
                /// </summary>
                /// <returns></returns>
                public List<TSimpleSequence> GetSequences()
                {
#if DEBUG
                    System.Diagnostics.Stopwatch sw = new System.Diagnostics.Stopwatch();
                    sw.Start();
#endif

                    sequences.Clear();
                    byte by;
                    string name = null;
                    string comment = null;
                    ISequenceContainer sequence = seqTemplate.NewSequence();
                    bool readName=true;

                    

                    //As long as the buffer can still hold sequences
                    while (sequences.Count < bufferSize)
                    {
                        if (readName) //If this is a new sequence -> read the first 8000 bytes
                        {

                            StringBuilder sb = new StringBuilder("");
                            
                            //The end of the file is a int with the value -1
                            //can only be obtained if one byte is read
                            int peek = fs.ReadByte();
                            int byt = 0;
                            if (peek == -1)
                            {
                                break;
   
                            }
                            else if (peek == (byte)254)
                            {
                                name = "";
                            } 
                            else //if the obtained character was not the end of the file insert it into the byte array
                            {
                                sb.Append((char)peek);
                                while ((byt = fs.ReadByte()) != (byte)254)
                                {
                                    sb.Append((char)byt);
                                }


                            }
                            name = sb.ToString();
                            sb = new StringBuilder("");

                            while ((byt = fs.ReadByte()) != (byte)254)
                            {
                                sb.Append((char)byt);
                            }
                            if(sb.Length>0) comment = sb.ToString();

                            readName = false;


                        } 
                        else
                        {
                            by = (byte)fs.ReadByte();

                            if (by == (byte)255) //if the read character ist the sequence terminating character store the sequence 
                            {
                                //Readjust the buffer and the buffersize automatically; if the buffer has ever shrunken to 1 -> do not increase it anymore
                                if (sequence.Length > 2000000 && bufferSize > 1) bufferSize = 1;
                                else if (sequence.Length > 200000 && bufferSize > 10) bufferSize = 10;

                                if (validator.IsValid(sequence)) sequences.Add(templateSimpleSequence.NewInstance(name, comment, sequence));
                                else if (!ignoreInvalid) throw new InvalidSequenceException(validator);

                                
                                name = null;
                                comment = null;
                                readName = true;
                                sequence = seqTemplate.NewSequence();

                            }
                            else sequence.Append(by); //if the read byte is not sequence terminating append it to the sequence

                        }
                    }



#if DEBUG
                    sw.Stop();
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} ByteReader_Sequence; {1} files loaded in {2} ms; First file contained {3} nucleotides; Sequence used: {4}", DateTime.Now.TimeOfDay, sequences.Count, sw.ElapsedMilliseconds, sequences.Count <= 0 ? 0 : sequences[0].Length, seqTemplate.GetType().ToString()));


#endif




                    return sequences;
                }

                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return this.seqTemplate;
                    }
                    set
                    {
                        this.seqTemplate = value;
                    }
                }


    


                public void Close()
                {
                    this.fs.Close();
                }

       
            }





            /// <summary>
            /// Decorater for SequenceReader derived classes,
            /// makes sure that each SimpleSequence has an unique identifier (i.e. Fasta name)
            /// </summary>
            public class SequenceReaderDecorator_UniqueSeqID<TSimpleSequence>:ISequenceReader<TSimpleSequence>
                where TSimpleSequence: ISimpleSequence,IFastaFileReadable<TSimpleSequence>,new()
            {
                private ISequenceReader<TSimpleSequence> reader;
                private Dictionary<string, int> dict;
                public SequenceReaderDecorator_UniqueSeqID(ISequenceReader<TSimpleSequence> reader)
                {

                    dict = new Dictionary<string, int>();
                    this.reader = reader;
                }

                public List<TSimpleSequence> GetSequences()
                {
                    List<TSimpleSequence> temp = reader.GetSequences();
                    int countRedundant = 0;
                    foreach (TSimpleSequence seq in temp)
                    {
                        if (dict.ContainsKey(seq.Name))
                        {
                            string old = seq.Name;
                            countRedundant++;
                            int add = dict[seq.Name]++;
                            seq.Name = seq.Name + '_' + add.ToString();

                            //If the new sequence name is per chance already a entry in the dictionary, just add an star until the ID is unique
                            while (dict.ContainsKey(seq.Name)) 
                            {
                                seq.Name += "*";
                            }
                            dict.Add(seq.Name, 1);



                        }
                        else
                        {
                            dict.Add(seq.Name, 1);
                        }
                    }


                    System.Diagnostics.Debug.WriteLine(String.Format("{0} NucleotideSequenceReaderDecorator_UniqueSeqID; {1} files checked for identical ID; {2} files with identical ID identified", DateTime.Now.TimeOfDay, temp.Count, countRedundant));

                    return temp;
                }

                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return reader.SequencePrototype;
                    }
                    set
                    {
                        reader.SequencePrototype = value;
                    }
                }

                public void Close()
                {
                    this.reader.Close();
                }
            }



            /// <summary>
            /// Decorater for SequenceReader derived classes,
            /// makes sure that each Sequence is uppercase, converts each sequence into uppercase characters
            /// </summary>
            public class SequenceReaderDecorator_AssureUppercaseSequence<TSimpleSequence> : ISequenceReader<TSimpleSequence>
                where TSimpleSequence : ISimpleSequence, IFastaFileReadable<TSimpleSequence>, new()
            {
                private ISequenceReader<TSimpleSequence> reader;

                public SequenceReaderDecorator_AssureUppercaseSequence(ISequenceReader<TSimpleSequence> reader)
                {

                    this.reader = reader;
                }

                public List<TSimpleSequence> GetSequences()
                {
                    List<TSimpleSequence> temp = reader.GetSequences();
                    foreach (TSimpleSequence seq in temp)
                    {
                        seq.Sequence.ToUpper();
                    }


                    System.Diagnostics.Debug.WriteLine(String.Format("{0} SequenceReaderDecorator_AssureUppercaseSequence; {1} files checked and converted into uppercase equences ID", DateTime.Now.TimeOfDay, temp.Count));

                    return temp;
                }

                public void Close()
                {
                    this.reader.Close();
                }

                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return reader.SequencePrototype;
                    }
                    set
                    {
                        reader.SequencePrototype = value;
                    }
                }
            }
          



            /// <summary>
            /// Decorator for SequenceReader, extracts for a given collection of IPositionable elements the flanking sequences
            /// May be used in batch mode, reads a sequence and extracts the flanking sequences from the read sequence
            /// Returns an empty collection if no more sequences can be read.
            /// </summary>
            /// <typeparam name="TSimpleSequence"></typeparam>
            public class SequenceReaderDecorator_ExtractFlankingSequences<TSimpleSequence>:ISequenceReader<TSimpleSequence>
                where TSimpleSequence: ISimpleSequence, IPositionable, IFastaFileReadable<TSimpleSequence>,new()
            {
                private ISequenceReader<TSimpleSequence> reader;
                private bool individualFlankMode;
                private int flankUpstream=0;
                private int flankDownstream=0;
                private int totalFlankLength=0;
                private bool addPositions=true;
                private TSimpleSequence templateSimpleSequence = new TSimpleSequence();

                private ISequenceContainer template;
                private Dictionary<string, List<IPositionable>> toFlank;
                /// <summary>
                /// Extracts flanking sequences for a given collection of elements implementing IPositionable (e.g.: SSRs)
                /// Identification of the sequence in which to search the flanking sequence is done through the fasta ID, so make sure that the .ParentName of the IPositionable fits the Name of the fasta ID. 
                /// Therefore decorators changing the fasta IDs should not be used prior to the extraction process.
                /// The flanking length upstream (5') and downstream (3') of the IPositionable element have to be specified, thus flanking sequences will have variing length. Dependend on the length of the IPositionable element
                /// The total flanking length might be less than the sequence which to flank
                /// </summary>
                public SequenceReaderDecorator_ExtractFlankingSequences(ISequenceReader<TSimpleSequence> reader, List<IPositionable> extractFrom, int flankUpstream, int flankDownstream)
            {
                individualFlankMode = true;
                this.reader = reader;
                toFlank = GroupPositionablesWithEqualParent(extractFrom);
                this.flankDownstream = flankDownstream;
                this.flankUpstream = flankUpstream;
                template = SequenceFactory.GetMediumSequence();
            }

                /// <summary>
                /// Extracts flanking sequences for a given collection of elements implementing IPositionable (e.g.: SSRs)
                /// Identification of the sequence in which to search the flanking sequence is done through the fasta ID, so make sure that the .ParentName of the IPositionable fits the Name of the fasta ID. 
                /// Therefore decorators changing the fasta IDs should not be used prior to the extraction process.
                /// The total length of the flanking sequence has to be specified, thus flanking sequences will always have the same length
                /// The total flanking length might be less than the sequence which to flank
                /// </summary>
                public SequenceReaderDecorator_ExtractFlankingSequences(ISequenceReader<TSimpleSequence> reader, List<IPositionable> extractFrom, int totalFlankLength)
                {
                    individualFlankMode = false;
                    this.totalFlankLength = totalFlankLength;
                    this.reader = reader;
                    toFlank = GroupPositionablesWithEqualParent(extractFrom);
                    template = SequenceFactory.GetMediumSequence();
                }

                /// <summary>
                /// Extracts flanking sequences for a given collection of elements implementing IPositionable (e.g.: SSRs)
                /// Identification of the sequence in which to search the flanking sequence is done through the fasta ID, so make sure that the .ParentName of the IPositionable fits the Name of the fasta ID. 
                /// Therefore decorators changing the fasta IDs should not be used prior to the extraction process.
                /// The flanking length upstream (5') and downstream (3') of the IPositionable element have to be specified, thus flanking sequences will have variing length. Dependend on the length of the IPositionable element
                /// The total flanking length might be less than the sequence which to flank
                /// </summary>
                public SequenceReaderDecorator_ExtractFlankingSequences(ISequenceReader<TSimpleSequence> reader, List<IPositionable> extractFrom,bool addPositions, int flankUpstream, int flankDownstream):this(reader,extractFrom,flankUpstream,flankDownstream)
                {
                    this.addPositions = addPositions;

                }

                /// <summary>
                /// Extracts flanking sequences for a given collection of elements implementing IPositionable (e.g.: SSRs)
                /// Identification of the sequence in which to search the flanking sequence is done through the fasta ID, so make sure that the .ParentName of the IPositionable fits the Name of the fasta ID. 
                /// Therefore decorators changing the fasta IDs should not be used prior to the extraction process.
                /// The total length of the flanking sequence has to be specified, thus flanking sequences will always have the same length
                /// The total flanking length might be less than the sequence which to flank
                /// </summary>
                public SequenceReaderDecorator_ExtractFlankingSequences(ISequenceReader<TSimpleSequence> reader, List<IPositionable> extractFrom, bool addPositions, int totalFlankLength)
                    : this(reader, extractFrom, totalFlankLength)
                {
                    this.addPositions = addPositions;

                }


                public List<TSimpleSequence> GetSequences()
                {
                    List<TSimpleSequence> extractFrom;

                    List<TSimpleSequence> extracted = new List<TSimpleSequence>();
                    do
                    {
                        extractFrom = reader.GetSequences();
                        if (extractFrom.Count < 1) return extractFrom;//If no more sequences can be read return the empty list


                        foreach (INucleotideSequence parent in extractFrom)
                        {
                            //If the Hash table contains the appropriate key, do:
                            if (toFlank.ContainsKey(parent.Name))
                            {
                                List<IPositionable> childs = toFlank[parent.Name];
                                foreach (IPositionable child in childs)
                                {
                                    TSimpleSequence flank;
                                    int start;
                                    int end;
                                    int length;
                                    string addToName = "";





                                    //Calculate the start position and the length, in dependency of the used flanking mode
                                    if (!individualFlankMode)
                                    {
                                        length = totalFlankLength;
                                        start = child.Start.Value - ((totalFlankLength - (child.End.Value - child.Start.Value + 1)) / 2) - 1;

                                    }
                                    else
                                    {
                                        length = child.End.Value - child.Start.Value + 1 + flankDownstream + flankUpstream;
                                        start = child.Start.Value - flankUpstream - 1;

                                        //123 456 childstart=6 flank upstream=2
                                        //TTT TTACACACAC
                                        //012 3  start=3;
                                    }


                                    if (length > parent.Length)
                                    {
                                        start = 0;
                                        length = (int)parent.Length;
                                    }
                                    else
                                    {

                                        if (start < 0) start = 0;
                                        if (start + length > parent.Length)
                                        {
                                            //01234567890123 L_14
                                            //TTTTTGTGTGTGTG
                                            start = ((int)parent.Length) - length;
                                        }

                                    }

                                    end = start + length - 1;


                                    int startToFlank_Child = 0;
                                    int endToFlank_Child = 0;
                                    int featureLength = child.End.Value - child.Start.Value + 1;

                                    //123 456 child.start=6 flank upstream=2 start=3
                                    //TTT TTACACACAC
                                    //--- 0123 startToFlan=2 = child.start-start-1

                                    //123456 childstart=1;
                                    //ACACACA
                                    //0123456 startToflank=0;
                                    startToFlank_Child = child.Start.Value - start - 1;

                                    //TTTTTAAAAATTTTT
                                    //01234567890
                                    endToFlank_Child = startToFlank_Child + featureLength - 1;

                                    if (addPositions) addToName = String.Format("  Startpos_in_parent={0}  Startpos_here={1}  Length={2}",child.Start.Value, startToFlank_Child+1,featureLength);


                                    flank = templateSimpleSequence.NewInstance(child.ParentName+addToName, child.ParentName, template.NewSequence(parent.Sequence.ToString((int)start,(int)length)), start + 1, end + 1);

                                    extracted.Add(flank);





                                }
                            }

                        }
                    } while (extracted.Count < 1); //as long as extracted does not contain sequences continue with the procedure

                    return extracted;


                }

                public void Close()
                {
                    this.reader.Close();
                }

                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                       return template;
                    }
                    set
                    {
                        this.template = value;
                    }
                }

                public static Dictionary<string, List<IPositionable>> GroupPositionablesWithEqualParent(List<IPositionable> positionables)
                {
                    Dictionary<string, List<IPositionable>> groupedSeq = new Dictionary<string, List<IPositionable>>();
                    foreach (IPositionable p in positionables)
                    {
                        if (groupedSeq.ContainsKey(p.ParentName))
                        {
                            groupedSeq[p.ParentName].Add(p);
                        }
                        else
                        {
                            List<IPositionable> tempList = new List<IPositionable>();
                            tempList.Add(p);
                            groupedSeq.Add(p.ParentName,tempList);
                        }
                    }
                    return groupedSeq;
                }


            }



            /// <summary>
            /// Decorator for SequenceReader, masks the sequence tracts corresponding to a given collection of IPositionable with the specified character
            /// </summary>
            /// <typeparam name="TSimpleSequence"></typeparam>
            public class SequenceReaderDecorator_MaskSequence<TSimpleSequence> : ISequenceReader<TSimpleSequence>
                where TSimpleSequence : ISimpleSequence, IPositionable, IFastaFileReadable<TSimpleSequence>, new()
            {
                private ISequenceReader<TSimpleSequence> reader;
                private TSimpleSequence templateSimpleSequence = new TSimpleSequence();
                private Dictionary<string, List<IPositionable>> toMask;
                private char mask = 'N';

                /// <summary>
                /// the positionables in the collection toMask act as templates for masking the corresponding sequence-tracts in the SimpleSequences passed by the reader
                /// </summary>
                public SequenceReaderDecorator_MaskSequence(ISequenceReader<TSimpleSequence> reader, List<IPositionable> toMask, char mask)
                {

                    this.reader = reader;
                    this.toMask = SequenceReaderDecorator_ExtractFlankingSequences<NucleotideSequence>.GroupPositionablesWithEqualParent(toMask);
                    this.mask = mask;
                }

                public SequenceReaderDecorator_MaskSequence(ISequenceReader<TSimpleSequence> reader, List<IPositionable> toMask)
                {

                    this.reader = reader;
                    this.toMask = SequenceReaderDecorator_ExtractFlankingSequences<NucleotideSequence>.GroupPositionablesWithEqualParent(toMask);

                }



                public List<TSimpleSequence> GetSequences()
                {
                    List<TSimpleSequence> sequences =reader.GetSequences();
                    List<TSimpleSequence> returnSequences = new List<TSimpleSequence>(sequences.Count);
                    foreach (TSimpleSequence s in sequences)
                    {
                        ISequenceContainer m = s.Sequence;
                        int countPositionables = 0;
                        if (toMask.ContainsKey(s.Name))
                        {
                            List<IPositionable> positionable = toMask[s.Name];
                            countPositionables = positionable.Count;

                            foreach (IPositionable pos in positionable)
                            {
                                for (int i = pos.Start.Value; i <= pos.End.Value; i++)
                                {
                                    m[i - 1] = mask;
                                }
                            }

                        }
                        string comment = s.Comment ?? "";
                        comment += String.Format(" {0} Positioables have been masked", countPositionables);
                        returnSequences.Add(templateSimpleSequence.NewInstance(s.Name, comment, m));

                    }


                    return returnSequences;

                }


                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return reader.SequencePrototype;
                    }
                    set
                    {
                        this.reader.SequencePrototype = value;
                    }
                }

                public void Close()
                {
                    this.reader.Close();
                }
            }



            /// <summary>
            /// Decorater for SequenceReader derived classes,
            /// makes sure that each Sequence is uppercase, converts each sequence into uppercase characters
            /// </summary>
            public class NucleotideSequenceReaderDecorator_ObtainQualityScores<TNucleotideSequence> : ISequenceReader<TNucleotideSequence>
                where TNucleotideSequence : INucleotideSequence
  
            {
                private ISequenceReader<TNucleotideSequence> reader;
                private string[] qualFiles;
                private int bufferQualityFiles;
                private bool ignoreEmptyQuality = false;

                public NucleotideSequenceReaderDecorator_ObtainQualityScores(ISequenceReader<TNucleotideSequence> reader, string[] qualityFiles)
                {

                    this.reader = reader;
                    this.qualFiles = qualityFiles;
                }

                public NucleotideSequenceReaderDecorator_ObtainQualityScores(ISequenceReader<TNucleotideSequence> reader, string[] qualityFiles, int bufferQualityFiles)
                {
                    this.reader = reader;
                    this.qualFiles = qualityFiles;
                    this.bufferQualityFiles = bufferQualityFiles;
                }

                public NucleotideSequenceReaderDecorator_ObtainQualityScores(ISequenceReader<TNucleotideSequence> reader, string[] qualityFiles, int bufferQualityFiles, bool ignoreEmptyQuality)
                {
                    this.reader = reader;
                    this.qualFiles = qualityFiles;
                    this.bufferQualityFiles = bufferQualityFiles;
                    this.ignoreEmptyQuality = ignoreEmptyQuality;
                }

                public List<TNucleotideSequence> GetSequences()
                {
                    List<TNucleotideSequence> temp = reader.GetSequences();

                    Dictionary<string, TNucleotideSequence> nsDict = new Dictionary<string, TNucleotideSequence>(temp.Count);
                    foreach (TNucleotideSequence ins in temp)
                    {
                        if (!nsDict.ContainsKey(ins.Name))
                        {
                            nsDict.Add(ins.Name, ins);
                        }
                        else throw new InvalidDataException("No two fasta sequences are permited to share the same name (fasta ID)");
                    }

                    for (int i = 0; i < qualFiles.Length; i++)
                    {
                        StreamReader sr = new StreamReader(qualFiles[i], Encoding.ASCII);
                        QualityFileReader qr = new QualityFileReader(sr, bufferQualityFiles);
                        List<QualitySequence> tempQual;
                        while ((tempQual = qr.GetSequences()).Count > 0)
                        {
                            foreach (QualitySequence qs in tempQual)
                            {
                                if (nsDict.ContainsKey(qs.Name)) nsDict[qs.Name].QualitySequence = qs;
                            }
                        }
                        tempQual.Clear();
                    }

                    if (!ignoreEmptyQuality)
                    {
                        foreach (TNucleotideSequence seq in temp)
                        {
                            if (seq.QualitySequence == null) throw new ArgumentNullException("Quality sequence of {0} is empty; Make sure that the quality sequence IDs and the fasta IDs are identical",seq.Name);
                        }
                    }
                    return temp;
                }

                public void Close()
                {
                    this.reader.Close();
                }

                public ISequenceContainer SequencePrototype
                {
                    get
                    {
                        return reader.SequencePrototype;
                    }
                    set
                    {
                        reader.SequencePrototype = value;
                    }
                }
            }
          




            namespace Misc
            {


                public class FastaMultiFileReader_Sequence<TSimpleSequence> : ISequenceReader<TSimpleSequence> where TSimpleSequence: ISimpleSequence, IFastaFileReadable<TSimpleSequence>,new()
                {
                    private ISequenceContainer template=SequenceFactory.GetDefaultSequence();
                    private ISequenceValidator validator;
                    private int bufferSize=5000;
                    private bool ignoreInvalid;
                    private StreamReader[] srs;
                    private int fileNumber = 0;
                    private FastaReader_Sequence<TSimpleSequence> activeReader;



                    public FastaMultiFileReader_Sequence(string[] paths, ISequenceValidator validator, bool ignoreInvalid, int bufferSize) 
                    {
                        if (paths.Length < 1) throw new InvalidDataException("paths (string[] does not contain a valid path");
                        srs = new StreamReader[paths.Length];

                        for (int i = 0; i < paths.Length; i++)
                        {
                            srs[i] = new StreamReader(paths[i], Encoding.ASCII);
                        }
                        this.validator = validator;
                        this.ignoreInvalid = ignoreInvalid;
                        this.bufferSize = bufferSize;

                        this.NewReader();
                    }
                    public FastaMultiFileReader_Sequence(StreamReader[] srs, ISequenceValidator validator, bool ignoreInvalid, int bufferSize)
                     
                    {
                        if (srs.Length < 1) throw new InvalidDataException("srs (StreamReader[]) does not contain a valid StreamReader");
                        this.srs = srs;
                        this.validator = validator;
                        this.ignoreInvalid = ignoreInvalid;
                        this.NewReader();
                    }




                    public FastaMultiFileReader_Sequence(string[] paths, ISequenceValidator validator, bool ignoreInvalid)
                    {
                        if (paths.Length < 1) throw new InvalidDataException("paths (string[] does not contain a valid path");
                        srs= new StreamReader[paths.Length];

                       for(int i =0; i<paths.Length;i++)
                        {
                            srs[i] = new StreamReader(paths[i], Encoding.ASCII);
                        }

                        this.validator = validator;
                        this.ignoreInvalid = ignoreInvalid;


                        this.NewReader();

                    }


                    public FastaMultiFileReader_Sequence(StreamReader[] srs, ISequenceValidator validator, bool ignoreInvalid)
                      
                    {
                        if (srs.Length < 1) throw new InvalidDataException("srs (StreamReader[]) does not contain a valid StreamReader");
                        this.srs = srs;
                        this.validator = validator;
                        this.ignoreInvalid = ignoreInvalid;
                        this.bufferSize = 5000;

                        this.NewReader();
                    }



                    public List<TSimpleSequence> GetSequences()
                    {
                        List<TSimpleSequence> temp;
                        while ((temp = activeReader.GetSequences()).Count < 1)
                        {
                            srs[fileNumber].Close(); //Close old StreamReader

                            fileNumber++;
                            if (fileNumber < srs.Length)
                            {
                                NewReader();
                            }
                            else
                            {
                                return temp;
                            }
                        }
                        return temp;
                    }



                    private void NewReader()
                    {
                        activeReader = new FastaReader_Sequence<TSimpleSequence>(srs[fileNumber], validator, ignoreInvalid, bufferSize);
                        this.activeReader.SequencePrototype = this.template;
                    }

                    public ISequenceContainer SequencePrototype
                    {
                        get
                        {
                            return this.template;
                        }
                        set
                        {
                            this.template=value;
                            activeReader.SequencePrototype = value;
                        }
                    }

                    public void Close()
                    {
                        this.activeReader.Close();
                    }
                }


                public class FastaMultiFileReader_Accumulator<TSimpleSequence> : ISequenceReader<TSimpleSequence> where TSimpleSequence : ISimpleSequence, IFastaFileReadable<TSimpleSequence>, new()
                {
                    private ISequenceContainer template = SequenceFactory.GetDefaultSequence();
                    private ISequenceValidator validator;
                    private bool ignoreInvalid;
                    private bool hasRead=false;
                    private string[] paths;




                    public FastaMultiFileReader_Accumulator(string[] paths, ISequenceValidator validator, bool ignoreInvalid)
                    {
                        if (paths.Length < 1) throw new InvalidDataException("paths (string[] does not contain a valid path");
     
                        this.paths=paths;
                        this.validator = validator;
                        this.ignoreInvalid = ignoreInvalid;
                    }




                    public List<TSimpleSequence> GetSequences()
                    {
                        List<TSimpleSequence> ret=new List<TSimpleSequence>();
                        if (!hasRead)
                        {
                            foreach (string s in paths)
                            {
                                List<TSimpleSequence> temp;
                                StreamReader sr = new StreamReader(s);
                                FastaReader_Sequence<TSimpleSequence> activeReader = new FastaReader_Sequence<TSimpleSequence>(sr, this.validator, this.ignoreInvalid);
                                activeReader.SequencePrototype = this.template;

                                while ((temp = activeReader.GetSequences()).Count >0)
                                {
                                    ret.AddRange(temp);
                                    temp.Clear();
   
                                }
                                sr.Close();
                            }

                            hasRead = true;
                        }
                        return ret;
                    }




                    /// <summary>
                    /// Get or set the prototype
                    /// </summary>
                    public ISequenceContainer SequencePrototype
                    {
                        get
                        {
                            return this.template;
                        }
                        set
                        {
                            this.template = value;
          
                        }
                    }

                    public void Close()
                    {
                    }
                }



                public class FastasReader_NucSeqInfo: INucSeqInfoReader
                {
                    private INucSeqInfoReader activeReader;
                    private ISequenceValidator validator;
                    private int bufferSize=5000;
                    private bool ignoreInvalid;
                    private StreamReader[] srs;
                    private bool fullDetails=true;
                    private int fileNumber = 0;

  
                                    /// <summary>
                /// Constructor for reading only the fasta ID, the length and the fasta comment,
                /// all fasta sequences are valid
                /// </summary>
                public FastasReader_NucSeqInfo(string[] paths)
                {
                    fullDetails=false;
                    srs= new StreamReader[paths.Length];
                    for(int i=0;i<paths.Length;i++)
                    {
                        srs[i]=new StreamReader(paths[i],Encoding.ASCII);
                    }
                    this.NewReader();
                }

                /// <summary>
                /// Constructor for reading only the fasta ID, the length and the fasta comment,
                /// all fasta sequences are valid
                /// </summary>
                public FastasReader_NucSeqInfo(StreamReader[] srs)
                {
                    this.fullDetails = false;
                    this.srs=srs;
                    this.NewReader();
                }




                /// <summary>
                /// Constructor for reading only the fasta ID, the length and the fasta comment,
                /// all fasta sequences are valid
                /// </summary>
                public FastasReader_NucSeqInfo(string[] paths, int bufferSize)
                {
                    this.fullDetails = false;
                    srs= new StreamReader[paths.Length];
                    this.bufferSize=bufferSize;
                    for(int i=0;i<paths.Length;i++)
                    {
                        srs[i]=new StreamReader(paths[i],Encoding.ASCII);
                    }
                    this.NewReader();
                }

                /// <summary>
                /// Constructor for reading only the fasta ID, the length and the fasta comment,
                /// all fasta sequences are valid
                /// </summary>
                public FastasReader_NucSeqInfo(StreamReader[] srs,int bufferSize)
                {
                    this.fullDetails = false;
                    this.srs=srs;
                    this.bufferSize=bufferSize;
                    this.NewReader();
                }



                public FastasReader_NucSeqInfo(string[] paths, ISequenceValidator validator, bool ignoreInvalid, int bufferSize)
                {                  

                    srs= new StreamReader[paths.Length];
                    this.bufferSize=bufferSize;
                    for(int i=0;i<paths.Length;i++)
                    {
                        srs[i]=new StreamReader(paths[i],Encoding.ASCII);
                    }
                    this.validator=validator;
                    this.ignoreInvalid=ignoreInvalid;
                    this.NewReader();
                
                
                }
                public FastasReader_NucSeqInfo(StreamReader[] srs, ISequenceValidator validator, bool ignoreInvalid, int bufferSize)
                {
                    this.srs = srs;
                    this.validator = validator;
                    this.bufferSize = bufferSize;
                    this.ignoreInvalid = ignoreInvalid;
                    this.NewReader();
                }

                public FastasReader_NucSeqInfo(string[] paths, ISequenceValidator validator, bool ignoreInvalid)
                {
                    srs = new StreamReader[paths.Length];
                    for (int i = 0; i < paths.Length; i++)
                    {
                        srs[i] = new StreamReader(paths[i], Encoding.ASCII);
                    }
                    this.validator = validator;
                    this.ignoreInvalid = ignoreInvalid;
                    this.NewReader();

                }
                public FastasReader_NucSeqInfo(StreamReader[] srs, ISequenceValidator validator, bool ignoreInvalid)

                {
                    this.srs = srs;
                    this.validator = validator;
                    this.ignoreInvalid = ignoreInvalid;
                    this.NewReader();
                }




                    public List<INucSeqInfo> GetSequences()
                    {
                        List<INucSeqInfo> temp;
                        while ((temp = activeReader.GetSequences()).Count < 1)
                        {
                            srs[fileNumber].Close(); //Close old StreamReader

                            fileNumber++;
                            if (fileNumber < srs.Length)
                            {
                                NewReader();
                            }
                            else
                            {
                                return temp;
                            }
                        }
                        return temp;
                    }

                    private void NewReader()
                    {
                        if (fullDetails) activeReader = new FastaReader_NucSeqInfo(srs[fileNumber], validator, ignoreInvalid, bufferSize);
                        else activeReader = new FastaReader_NucSeqInfo(srs[fileNumber], bufferSize);
                    }

                    public void Close()
                    {
                        this.activeReader.Close();
                    }
                }


            }
        }
        namespace Repetitive
        {

            namespace Microsat
            {
                using Bio.Repetitive.Microsat;


                public interface ISSRWriter
                {
                    void WriteInfo(OverallInfo searchInfo);
                    void WriteSSRs(List<SSR> ssrs);
                }

                public interface ISSRReader
                {
                    List<SSR> ReadSSRs();
                    bool IsInfoReady { get;}
                    OverallInfo ReadInfo();

                }



                /// <summary>
                /// Writes a collection of SSRCouples into a file;
                /// </summary>
                public class SSRCoupleWriter
                {
                    private StreamWriter sw;
                    public SSRCoupleWriter(StreamWriter sw)
                    {
                        this.sw = sw;
                    }

                    public void WriteCouples(List<SSRCouple> couples)
                    {
                        //First write the definition to the file
                        sw.WriteLine("//");
                        sw.WriteLine("// SSR_Number:\t" + SSR.ToStringHeader());
                        sw.WriteLine("//");
                        sw.WriteLine();
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif

                        //Than write each single SSR couple to the file
                        foreach (SSRCouple c in couples)
                        {
                            sw.WriteLine("1:\t" + c.First.ToString());
                            sw.WriteLine("2:\t " + c.Second.ToString());
                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRCoupleWriter; Wrote {1} SSR couples into output file in {2} ms", DateTime.Now.TimeOfDay, couples.Count, stopw.ElapsedMilliseconds));
#endif

                    }

                    public void WriteInfo(OverallInfo info)
                    {

                        sw.WriteLine();
                        sw.WriteLine("#Nucleotide count: {0}", info.nucleotideCount);
                        sw.WriteLine("#File count: {0}", info.fileCount);
                        sw.WriteLine();

                    }


                }

                /// <summary>
                /// Reads SSRCouples from a given SSR-Couple file
                /// </summary>
                public class SSRCoupleReader
                {
                    private StreamReader sr;
                    private readonly int bufferSize;
                    private bool infoReady = false;
                    private OverallInfo searchInfo;



                    public SSRCoupleReader(StreamReader sr, int bufferSize)
                    {
                        this.sr = sr;
                        this.bufferSize = bufferSize;

                    }
                    public SSRCoupleReader(StreamReader sr)
                    {
                        this.sr = sr;
                        this.bufferSize = 5000;

                    }

                    public List<SSRCouple> ReadCouples()
                    {
                        List<SSRCouple> couples = new List<SSRCouple>();
                        string line = "";
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif

                        SSR first = null;
                        SSR second = null;

                        //As long as there are lines in rhe file proceed!
                        while ((couples.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                        {
                            //Get rid of space
                            line = line.Trim();
                            if (line == "") continue;

                            //The > resets
                            if (line.StartsWith("//")) continue;
                            else if (line.StartsWith("1:"))
                            {
                                first = ParseLine(line);


                            }
                            else if (line.StartsWith("2:"))
                            {
                                second = ParseLine(line);
                                if (first == null || second == null) throw new Exception("Error during reading of the SSR couples");
                                couples.Add(new SSRCouple(first, second));
                                first = null;
                                second = null;
                            }
                            else if (line.StartsWith("#"))
                            {
                                int startPos = 0;
                                string substring = "";

                                if (line.Contains("Nucleotide"))
                                {
                                    startPos = line.IndexOf("count:") + 6;
                                    substring = line.Substring(startPos, line.Length - startPos);
                                    this.searchInfo.nucleotideCount = Convert.ToInt64(substring.Trim());
                                    infoReady = true;
                                }
                                else if (line.Contains("File"))
                                {
                                    startPos = line.IndexOf("count:") + 6;
                                    substring = line.Substring(startPos, line.Length - startPos);
                                    this.searchInfo.fileCount = Convert.ToInt32(substring.Trim());
                                    infoReady = true;

                                }

                            }

                        }
#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRCoupleReader; {1} SSR couples read in {2} ms", DateTime.Now.TimeOfDay, couples.Count, stopw.ElapsedMilliseconds));
#endif
                        return couples;

                    }

                    //Read the overall search info 
                    public OverallInfo ReadInfo()
                    {
                        if (!this.IsInfoReady) throw new ArgumentException("The overall info is not yet ready; Use the IsInfoReady property prior to the use of ReadInfo()");
                        return searchInfo;
                    }

                    //Extracts the SSR related information from the input string
                    private SSR ParseLine(string line)
                    {
                        string[] subs = line.Split(new char[] { '\t' });


                        //Def of SSR.ToString():  1:this.parentName, this.motif_Found, this.motif_FullyStandardized, 4:this.Start, this.End, 6:this.Length, this.score, 8: this.mismatches, this.NucleotideSequenc);

                        return new SSR(subs[1].Trim(), subs[2].Trim(), SequenceFactory.GetSmallSequence(subs[9].Trim()), Convert.ToInt32(subs[4].Trim()), Convert.ToInt32(subs[5].Trim()), Convert.ToInt32(subs[7].Trim()), Convert.ToInt32(subs[8].Trim()), SSRSearchMode.Unknown);

                    }

                    //Indicates whether the nucleotideInfo file is ready
                    public bool IsInfoReady
                    {
                        get
                        {
                            return infoReady;
                        }
                    }
                }


                public class SSRReaderSciRoKoParse : ISSRReader
                {
                    private StreamReader sr;
                    bool infoReady = false;
                    private readonly int bufferSize;
                    private OverallInfo searchInfo;
                    string nameOfSequence;
                    public SSRReaderSciRoKoParse(StreamReader sr, int bufferSize)
                    {
                        this.sr = sr;
                        this.bufferSize = bufferSize;
                        nameOfSequence = "";
                    }
                    public SSRReaderSciRoKoParse(StreamReader sr)
                    {
                        this.sr = sr;
                        this.bufferSize = 5000;
                        nameOfSequence = "";
                    }



                    /// <summary>
                    /// Reads from a SciRoKo-File
                    /// </summary>
                    /// <returns>A list full of SSRs</returns>
                    public List<SSR> ReadSSRs()
                    {

                        List<SSR> ssrList = new List<SSR>();
                        string line = "";
                        string sciRoKoInfo = "";
                        string ssrSequenze = "";
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        //As long as the Buffer is not full and there are lines in the file proceed!
                        while ((ssrList.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                        {
                            //Get rid of space
                            line = line.Trim();
                            if (line == "") continue;

                            //The > resets
                            if (line.StartsWith(">"))
                            {
                                if (nameOfSequence != "" && ssrSequenze != "" && sciRoKoInfo != "")
                                {

                                    ParseSciRoKoSSR(ssrList, nameOfSequence, sciRoKoInfo, ssrSequenze);
                                    nameOfSequence = "";
                                    ssrSequenze = "";
                                    sciRoKoInfo = "";
                                }
                                //After saving the previous SSR set the name of the new one.
                                nameOfSequence = line.Substring(1, line.Length - 1);

                            }
                            else if (line.Contains("Motif"))
                            {
                                sciRoKoInfo = line;
                            }
                            else if (line.StartsWith("#"))
                            {
                                int startPos = 0;
                                string substring = "";

                                if (line.Contains("Nucleotide"))
                                {
                                    startPos = line.IndexOf("count:") + 6;
                                    substring = line.Substring(startPos, line.Length - startPos);
                                    this.searchInfo.nucleotideCount = Convert.ToInt64(substring.Trim());
                                    infoReady = true;
                                }
                                else if (line.Contains("File"))
                                {
                                    startPos = line.IndexOf("count:") + 6;
                                    substring = line.Substring(startPos, line.Length - startPos);
                                    this.searchInfo.fileCount = Convert.ToInt32(substring.Trim());
                                    infoReady = true;
                                }

                            }
                            else ssrSequenze += line;

                        }


                        //Also store last file
                        if (nameOfSequence != "" && ssrSequenze != "" && sciRoKoInfo != "")
                        {
                            ParseSciRoKoSSR(ssrList, nameOfSequence, sciRoKoInfo, ssrSequenze);
                            nameOfSequence = "";
                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRReaderSciRoKo; {1} SSRs read in {2} ms", DateTime.Now.TimeOfDay, ssrList.Count, stopw.ElapsedMilliseconds));
#endif
                        return ssrList;

                    }

                    private void ParseSciRoKoSSR(List<SSR> ssrs, string nameOfSequence, string sciRoKoInfo, string ssrSequence)
                    {
                        //Determine motif
                        int tempPos1 = sciRoKoInfo.IndexOf("Motif:") + 6;
                        string subString = sciRoKoInfo.Substring(tempPos1, sciRoKoInfo.IndexOf("- SSR start:") - tempPos1);
                        string motif = subString.Trim();

                        //Check if motif fits the criteria if not, skip it
                        SequenceCharacterKind motifKind = NucleotideSequenceUtility.GetSequenceCharacterKind(motif);
                        if (motifKind != SequenceCharacterKind.ATCG) return;

                        //Determine SSRStart
                        tempPos1 = sciRoKoInfo.IndexOf("- SSR start:") + 12;
                        subString = sciRoKoInfo.Substring(tempPos1, sciRoKoInfo.IndexOf("- SSR end:") - tempPos1);
                        int ssrStart = Convert.ToInt32(subString.Trim());

                        //Determine SSREnd
                        tempPos1 = sciRoKoInfo.IndexOf("- SSR end:") + 10;
                        subString = sciRoKoInfo.Substring(tempPos1, sciRoKoInfo.IndexOf("- Total length:") - tempPos1);
                        int ssrEnd = Convert.ToInt32(subString.Trim());


                        //Determine TotalLength
                        tempPos1 = sciRoKoInfo.IndexOf("- Total length:") + 15;
                        subString = sciRoKoInfo.Substring(tempPos1, sciRoKoInfo.IndexOf("- Score:") - tempPos1);
                        int totalLength = Convert.ToInt32(subString.Trim());

                        //Determine Score
                        tempPos1 = sciRoKoInfo.IndexOf("- Score:") + 8;
                        subString = sciRoKoInfo.Substring(tempPos1, sciRoKoInfo.IndexOf("- Mismatches:") - tempPos1);
                        int score = Convert.ToInt32(subString.Trim());

                        //Determine Mismatches
                        tempPos1 = sciRoKoInfo.IndexOf("- Mismatches:") + 13;
                        subString = sciRoKoInfo.Substring(tempPos1, sciRoKoInfo.IndexOf("- Search Mode:") - tempPos1);
                        int mismatches = Convert.ToInt32(subString.Trim());

                        //Determine SearchMode
                        tempPos1 = sciRoKoInfo.IndexOf("- Search Mode:") + 14;
                        subString = sciRoKoInfo.Substring(tempPos1, sciRoKoInfo.Length - tempPos1);
                        string sMode = subString.Trim();

                        SSRSearchMode searchMode;
                        switch (sMode)
                        {
                            case "MisMatchFixedPenalty":
                                searchMode = SSRSearchMode.MisMatchFixedPenalty;
                                break;
                            case "MisMatchVariablePenalty":
                                searchMode = SSRSearchMode.MisMatchVariablePenalty;
                                break;
                            case "PerfectLength":
                                searchMode = SSRSearchMode.PerfectLength;
                                break;
                            case "PerfectRepeats":
                                searchMode = SSRSearchMode.PerfectRepeats;
                                break;
                            case "ModifiedSputnikI":
                                searchMode = SSRSearchMode.ModifiedSputnikI;
                                break;
                            case "ModifiedSputnikII":
                                searchMode = SSRSearchMode.ModifiedSputnikII;
                                break;
                            case "Sputnik":
                                searchMode = SSRSearchMode.Sputnik;
                                break;

                            default: searchMode = SSRSearchMode.Unknown;
                                break;

                        }

                        //Add the SSR
                        ssrs.Add(new SSR(nameOfSequence, motif, SequenceFactory.GetSmallSequence(ssrSequence), ssrStart, ssrEnd, score, mismatches, searchMode));


                    }


                    /// <summary>
                    /// Indicates whether the Info is accessible and can be read;
                    /// SciRoKo usually stores the overall search info at the end of the file, therefore generally the info will be read after all
                    /// SSRs have been read
                    /// </summary>
                    public bool IsInfoReady
                    {
                        get
                        {
                            return infoReady;
                        }
                    }

                    /// <summary>
                    /// Returns an overall info containing the number of nucleotides and files that have been searched for microsatellites
                    /// </summary>
                    ///<exception cref="ArgumentException">The overall info is not yet ready; Use the IsInfoReady property prior to the use of the ReadInfo()</exception>
                    public OverallInfo ReadInfo()
                    {
                        if (!this.IsInfoReady) throw new ArgumentException("The overall info is not yet ready; Use the IsInfoReady property prior to the use of ReadInfo()");
                        return searchInfo;
                    }

                }


                public class SSRReaderSciRoKo : ISSRReader
                {
              
                    private StreamReader sr;
                    bool infoReady = false;
                    private readonly int bufferSize;
                    private OverallInfo searchInfo;
                    string nameOfSequence;

                    //Create a Regex for the following input line; The Regex should use only explicit capture and the regex should be compiled
                    //The Regex is compiled only once as it is static and therefore accessible to all members. The parenthesis are capturing!
                    //Motif: A - SSR start: 11073 - SSR end: 11098 - Total length: 26 - Score: 26 - Mismatches: 0 - Search Mode: MisMatchFixedPenalty
                    private static System.Text.RegularExpressions.Regex sciroRegex = new System.Text.RegularExpressions.Regex(
                        @"^Motif: (?<motif>\w+) - SSR start: (?<start>\d+) - SSR end: (?<end>\d+) - Total length: (?<lenght>\d+) - Score: (?<score>\d+) - Mismatches: (?<mm>\d+) - Search Mode: (?<mode>\w+)\s*$"
                        , System.Text.RegularExpressions.RegexOptions.ExplicitCapture | System.Text.RegularExpressions.RegexOptions.Compiled);


                    private static DNASequenceValidator_ATCG val = new DNASequenceValidator_ATCG();

                    public SSRReaderSciRoKo(StreamReader sr, int bufferSize)
                    {
                        this.sr = sr;
                        this.bufferSize = bufferSize;
                        nameOfSequence = "";
                    }
                    public SSRReaderSciRoKo(StreamReader sr)
                    {
                        this.sr = sr;
                        this.bufferSize = 5000;
                        nameOfSequence = "";
                    }



                    /// <summary>
                    /// Reads from a SciRoKo-File
                    /// </summary>
                    /// <returns>A list full of SSRs</returns>
                    public List<SSR> ReadSSRs()
                    {

                        List<SSR> ssrList = new List<SSR>();
                        string line = "";
                        string sciRoKoInfo = "";
                        string ssrSequenze = "";
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        //As long as the Buffer is not full and there are lines in the file proceed!
                        while ((ssrList.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                        {
                            //Get rid of space
                            line = line.Trim();
                            if (line == "") continue;

                            //The > resets
                            if (line.StartsWith(">"))
                            {
                                if (nameOfSequence != "" && ssrSequenze != "" && sciRoKoInfo != "")
                                {

                                    ParseSciRoKoSSR(ssrList, nameOfSequence, sciRoKoInfo, ssrSequenze);
                                    nameOfSequence = "";
                                    ssrSequenze = "";
                                    sciRoKoInfo = "";
                                }
                                //After saving the previous SSR set the name of the new one.
                                nameOfSequence = line.Substring(1, line.Length - 1);

                            }
                            else if (line.Contains("Motif"))
                            {
                                sciRoKoInfo = line;
                            }
                            else if (line.StartsWith("#"))
                            {
                                int startPos = 0;
                                string substring = "";

                                if (line.Contains("Nucleotide"))
                                {
                                    startPos = line.IndexOf("count:") + 6;
                                    substring = line.Substring(startPos, line.Length - startPos);
                                    this.searchInfo.nucleotideCount = Convert.ToInt64(substring.Trim());
                                    infoReady = true;
                                }
                                else if (line.Contains("File"))
                                {
                                    startPos = line.IndexOf("count:") + 6;
                                    substring = line.Substring(startPos, line.Length - startPos);
                                    this.searchInfo.fileCount = Convert.ToInt32(substring.Trim());
                                    infoReady = true;
                                }

                            }
                            else ssrSequenze += line;

                        }


                        //Also store last file
                        if (nameOfSequence != "" && ssrSequenze != "" && sciRoKoInfo != "")
                        {
                            ParseSciRoKoSSR(ssrList, nameOfSequence, sciRoKoInfo, ssrSequenze);
                            nameOfSequence = "";
                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRReaderSciRoKo; {1} SSRs read in {2} ms", DateTime.Now.TimeOfDay, ssrList.Count, stopw.ElapsedMilliseconds));
#endif
                        return ssrList;

                    }

                    private void ParseSciRoKoSSR(List<SSR> ssrs, string nameOfSequence, string sciRoKoInfo, string ssrSequence)
                    {
                         
                        //Match the regex against the text specified in sciRoKoInfo
                        System.Text.RegularExpressions.Match m=sciroRegex.Match(sciRoKoInfo);

                        //In case no match has been found throw an exception
                        if (!m.Success) throw new InvalidDataException("Regular expression could not match the sciRoKoInfo");

                        //Retrieve the motif from the matches
                        string motif = m.Groups["motif"].Value;
                        //Check if motif fits the criteria if not, skip it
                        if (!val.IsValid(new Sequence_String(motif))) return;        

                        string sMode = m.Groups["mode"].Value;

                        SSRSearchMode searchMode;
                        switch (sMode)
                        {
                            case "MisMatchFixedPenalty":
                                searchMode = SSRSearchMode.MisMatchFixedPenalty;
                                break;
                            case "MisMatchVariablePenalty":
                                searchMode = SSRSearchMode.MisMatchVariablePenalty;
                                break;
                            case "PerfectLength":
                                searchMode = SSRSearchMode.PerfectLength;
                                break;
                            case "PerfectRepeats":
                                searchMode = SSRSearchMode.PerfectRepeats;
                                break;
                            case "ModifiedSputnikI":
                                searchMode = SSRSearchMode.ModifiedSputnikI;
                                break;
                            case "ModifiedSputnikII":
                                searchMode = SSRSearchMode.ModifiedSputnikII;
                                break;
                            case "Sputnik":
                                searchMode = SSRSearchMode.Sputnik;
                                break;

                            default: searchMode = SSRSearchMode.Unknown;
                                break;

                        }

                        //Add the SSR
                        ssrs.Add(new SSR(nameOfSequence, motif, SequenceFactory.GetSmallSequence(ssrSequence), Convert.ToInt32(m.Groups["start"].Value), Convert.ToInt32(m.Groups["end"].Value), Convert.ToInt32(m.Groups["score"].Value), Convert.ToInt32(m.Groups["mm"].Value), searchMode));


                    }


                    /// <summary>
                    /// Indicates whether the Info is accessible and can be read;
                    /// SciRoKo usually stores the overall search info at the end of the file, therefore generally the info will be read after all
                    /// SSRs have been read
                    /// </summary>
                    public bool IsInfoReady
                    {
                        get
                        {
                            return infoReady;
                        }
                    }

                    /// <summary>
                    /// Returns an overall info containing the number of nucleotides and files that have been searched for microsatellites
                    /// </summary>
                    ///<exception cref="ArgumentException">The overall info is not yet ready; Use the IsInfoReady property prior to the use of the ReadInfo()</exception>
                    public OverallInfo ReadInfo()
                    {
                        if (!this.IsInfoReady) throw new ArgumentException("The overall info is not yet ready; Use the IsInfoReady property prior to the use of ReadInfo()");
                        return searchInfo;
                    }

                }




                /// <summary>
                /// Reads sputnik files from a given file
                /// </summary>
                public class SSRReaderSputnik : ISSRReader
                {
                    private StreamReader sr;
                    private readonly int bufferSize;
                  //  private OverallInfo searchInfo;
                    string nameOfSequence;


                    public SSRReaderSputnik(StreamReader sr, int bufferSize)
                    {
                        this.sr = sr;
                        this.bufferSize = bufferSize;
                        nameOfSequence = "";
                    }
                    public SSRReaderSputnik(StreamReader sr)
                    {
                        this.sr = sr;
                        this.bufferSize = 5000;
                        nameOfSequence = "";
                    }

                    public List<SSR> ReadSSRs()
                    {
                        bool ignoreSpecialChars = true;

                        List<SSR> ssrList = new List<SSR>();
                        string line = "";
                        string sputnikInfo = "";
                        string SSRsequenze = "";
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif

                        //As long as there are lines in rhe file proceed!
                        while ((ssrList.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                        {
                            //Get rid of space
                            line = line.Trim();
                            if (line == "") continue;

                            //The > resets
                            if (line.StartsWith(">"))
                            {
                                if (nameOfSequence != "" && SSRsequenze != "" && sputnikInfo != "")
                                {

                                    ParseSputnikSSR(ssrList, nameOfSequence, sputnikInfo, SSRsequenze, ignoreSpecialChars);
                                    nameOfSequence = "";
                                    SSRsequenze = "";
                                    sputnikInfo = "";
                                }
                                nameOfSequence = line.Substring(1, line.Length - 1);

                            }
                            else if (line.Contains("nucleotide"))
                            {
                                if (nameOfSequence != "" && SSRsequenze != "" && sputnikInfo != "")
                                {

                                    ParseSputnikSSR(ssrList, nameOfSequence, sputnikInfo, SSRsequenze, ignoreSpecialChars);
                                    SSRsequenze = "";
                                    sputnikInfo = "";
                                }
                                sputnikInfo = line;
                            }
                            else SSRsequenze += line;

                        }
                        //Also store last file
                        if (nameOfSequence != "" && SSRsequenze != "" && sputnikInfo != "")
                        {
                            ParseSputnikSSR(ssrList, nameOfSequence, sputnikInfo, SSRsequenze, ignoreSpecialChars);
                            nameOfSequence = "";
                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRReaderSputnik; {1} SSRs read in {2} ms", DateTime.Now.TimeOfDay, ssrList.Count, stopw.ElapsedMilliseconds));
#endif
                        return ssrList;
                    }
                    /// <summary>
                    /// The Sputnik file format does not support the SciRoKo SSR-search info
                    /// therefore this property returns false;
                    /// </summary>
                    public bool IsInfoReady
                    {
                        get
                        {
                            return false;
                        }
                    }

                    public OverallInfo ReadInfo()
                    {
                        OverallInfo info;
                        info.fileCount = 0;
                        info.nucleotideCount = 0;
                        return info;
                    }


                    //Extracts the SSR related information from the sputnikInfo string and stores the SSR in the database
                    private static void ParseSputnikSSR(List<SSR> ssrList, string fastaName, string sputnikInfo, string SSRsequenze, bool ignoreSpecialChars)
                    {
                        //Determine the motif
                        string dttp = sputnikInfo.Substring(0, sputnikInfo.IndexOf("nucleotide"));
                        int motifLength = 0;
                        switch (dttp)
                        {
                            case "single ": motifLength = 1;
                                break;
                            case "di": motifLength = 2;
                                break;
                            case "tri": motifLength = 3;
                                break;
                            case "tetra": motifLength = 4;
                                break;
                            case "penta": motifLength = 5;
                                break;
                            case "hexa": motifLength = 6;
                                break;

                        }
                        string motif = SSRsequenze.Substring(0, motifLength);
                        SequenceCharacterKind motifKind = NucleotideSequenceUtility.GetSequenceCharacterKind(motif);

                        //Check if motif fits the criteria
                        if (motifKind != SequenceCharacterKind.ATCG && ignoreSpecialChars) return;
                        else if (motifKind != SequenceCharacterKind.ATCG && !ignoreSpecialChars) throw new Exception("SSR obtained from Sputnik contains invalid characters");

                        //Determine total length
                        int tempPos1 = sputnikInfo.IndexOf("length") + 6;
                        string subString = sputnikInfo.Substring(tempPos1, sputnikInfo.IndexOf("score") - tempPos1);
                        int totalLength = Convert.ToInt32(subString.Trim());

                        //Determine score
                        tempPos1 = sputnikInfo.IndexOf("score") + 5;
                        subString = sputnikInfo.Substring(tempPos1, sputnikInfo.Length - tempPos1);
                        int score = Convert.ToInt32(subString.Trim());

                        //Determine SSRStart
                        tempPos1 = sputnikInfo.IndexOf("tide") + 4;
                        subString = sputnikInfo.Substring(tempPos1, sputnikInfo.IndexOf(":") - tempPos1);
                        int ssrStart = Convert.ToInt32(subString.Trim());

                        //Determine SSREnd
                        tempPos1 = sputnikInfo.IndexOf(":") + 1;
                        subString = sputnikInfo.Substring(tempPos1, sputnikInfo.IndexOf("-") - tempPos1);
                        int ssrEnd = Convert.ToInt32(subString.Trim());

                        //Determine number of mismatches
                        int mismatches = (SSRsequenze.Length - motif.Length - score) / 6;

                        //An score adjustment is not necessary because when recalculating for the SSRWriterSputnik
                        //score=totalLength-motifLength-mm*6

                        //Add the Microsatellite
                        ssrList.Add(new SSR(fastaName, motif, SequenceFactory.GetSmallSequence(SSRsequenze), ssrStart, ssrEnd, score, mismatches, SSRSearchMode.Sputnik));
                    }

                }

                /// <summary>
                /// Reads SSRs from a given Modified Sputnik I file
                /// </summary>
                public class SSRReaderModifiedSputnik1 : ISSRReader
                {
                    private StreamReader sr;
                    private readonly int bufferSize;
                   // private OverallInfo searchInfo;
                    string nameOfSequence;


                    public SSRReaderModifiedSputnik1(StreamReader sr, int bufferSize)
                    {
                        this.sr = sr;
                        this.bufferSize = bufferSize;
                        nameOfSequence = "";
                    }
                    public SSRReaderModifiedSputnik1(StreamReader sr)
                    {
                        this.sr = sr;
                        this.bufferSize = 5000;
                        nameOfSequence = "";
                    }

                    public List<SSR> ReadSSRs()
                    {
                        bool ignoreSpecialChars = true;

                        List<SSR> ssrList = new List<SSR>();
                        string line = "";
                        string sputnikInfo = "";
                        string SSRsequenze = "";
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        //As long as there are lines in rhe file proceed!
                        while ((ssrList.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                        {
                            //Get rid of space
                            line = line.Trim();
                            if (line == "") continue;

                            //The > resets
                            if (line.StartsWith(">"))
                            {
                                if (nameOfSequence != "" && SSRsequenze != "" && sputnikInfo != "")
                                {

                                    ParseSputnikSSR(ssrList, nameOfSequence, sputnikInfo, SSRsequenze, ignoreSpecialChars);
                                    nameOfSequence = "";
                                    SSRsequenze = "";
                                    sputnikInfo = "";
                                }
                                nameOfSequence = line.Substring(1, line.Length - 1);

                            }
                            else if (line.Contains("nucleotide"))
                            {
                                if (nameOfSequence != "" && SSRsequenze != "" && sputnikInfo != "")
                                {

                                    ParseSputnikSSR(ssrList, nameOfSequence, sputnikInfo, SSRsequenze, ignoreSpecialChars);
                                    SSRsequenze = "";
                                    sputnikInfo = "";
                                }
                                sputnikInfo = line;
                            }
                            else SSRsequenze += line;

                        }
                        //Also store last file
                        if (nameOfSequence != "" && SSRsequenze != "" && sputnikInfo != "")
                        {
                            ParseSputnikSSR(ssrList, nameOfSequence, sputnikInfo, SSRsequenze, ignoreSpecialChars);
                            nameOfSequence = "";
                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRReaderModifiedSputnik1; {1} SSRs read in {2} ms", DateTime.Now.TimeOfDay, ssrList.Count, stopw.ElapsedMilliseconds));
#endif
                        return ssrList;
                    }

                    /// <summary>
                    /// The Modified Sputnik I file format does not support the SciRoKo SSR-search info
                    /// therefore this property returns false;
                    /// </summary>
                    public bool IsInfoReady
                    {
                        get
                        {
                            return false;
                        }
                    }

                    public OverallInfo ReadInfo()
                    {
                        OverallInfo info;
                        info.fileCount = 0;
                        info.nucleotideCount = 0;
                        return info;
                    }


                    //Extracts the SSR related information from the sputnikInfo string and stores the SSR in the database
                    private static void ParseSputnikSSR(List<SSR> ssrList, string fastaName, string sputnikInfo, string SSRsequenze, bool ignoreSpecialChars)
                    {

                        //Determine Motif
                        int tempPos1 = sputnikInfo.IndexOf("unit") + 4;
                        string subString = sputnikInfo.Substring(tempPos1, sputnikInfo.Length - tempPos1);
                        string motif = subString.Trim();

                        SequenceCharacterKind motifKind = NucleotideSequenceUtility.GetSequenceCharacterKind(motif);

                        //Check if motif fits the criteria
                        if (motifKind != SequenceCharacterKind.ATCG && ignoreSpecialChars) return;
                        else if (motifKind != SequenceCharacterKind.ATCG && !ignoreSpecialChars) throw new Exception("SSR obtained from Sputnik contains invalid characters");

                        //Determine total length
                        tempPos1 = sputnikInfo.IndexOf("length") + 6;
                        subString = sputnikInfo.Substring(tempPos1, sputnikInfo.IndexOf("score") - tempPos1);
                        int totalLength = Convert.ToInt32(subString.Trim());

                        //Determine score
                        tempPos1 = sputnikInfo.IndexOf("score") + 5;
                        subString = sputnikInfo.Substring(tempPos1, sputnikInfo.IndexOf("unit") - tempPos1);
                        int score = Convert.ToInt32(subString.Trim());

                        //Determine SSRStart
                        tempPos1 = sputnikInfo.IndexOf("tide") + 4;
                        subString = sputnikInfo.Substring(tempPos1, sputnikInfo.IndexOf(":") - tempPos1);
                        int ssrStart = Convert.ToInt32(subString.Trim());

                        //Determine SSREnd
                        tempPos1 = sputnikInfo.IndexOf(":") + 1;
                        subString = sputnikInfo.Substring(tempPos1, sputnikInfo.IndexOf("-") - tempPos1);
                        int ssrEnd = Convert.ToInt32(subString.Trim());

                        //Also modified sputnik has the default mismatch penalty 6
                        //Determine number of mismatches
                        int mismatches = (SSRsequenze.Length - motif.Length - score) / 6;

                        //An score adjustment is not necessary because when recalculating for the SSRWriterSputnik
                        //score=totalLength-motifLength-mm*6

                        //Add the Microsatellite
                        ssrList.Add(new SSR(fastaName, motif, SequenceFactory.GetSmallSequence(SSRsequenze), ssrStart, ssrEnd, score + motif.Length, mismatches, SSRSearchMode.ModifiedSputnikI));
                    }

                }

                /// <summary>
                /// Reads SSRs from a given Modified Sputnik II file
                /// </summary>
                public class SSRReaderModifiedSputnik2 : ISSRReader
                {
                    private StreamReader sr;
                    private readonly int bufferSize;
                  //  private OverallInfo searchInfo;
                    Queue<SSR> ssrs = new Queue<SSR>();
                   // private int counter;
                    private string firstLine;
                    private string secondLine;
                    private string sequence;



                    public SSRReaderModifiedSputnik2(StreamReader sr, int bufferSize)
                    {
                        this.sr = sr;
                        this.bufferSize = bufferSize;


                    }
                    public SSRReaderModifiedSputnik2(StreamReader sr)
                    {
                        this.sr = sr;
                        this.bufferSize = 5000;


                    }

                    public List<SSR> ReadSSRs()
                    {

                        string line = "";

                        firstLine = "";
                        secondLine = "";
                        sequence = "";
                      //  counter = 0;
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif



                        //As long as there are lines in rhe file proceed!
                        while ((ssrs.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                        {


                            //Get rid of space
                            line = line.Trim();
                            if (line == "" || line.StartsWith("#")) continue;

                            else if (line.Contains("nucleotide"))
                            {
                                if (firstLine == "") firstLine = line;
                                else secondLine = line;

                                if (secondLine != "")
                                {

                                    ParseSputnikSSR(ssrs);

                                    firstLine = secondLine;
                                    secondLine = "";
                                    sequence = "";
                                }


                            }
                            else
                            {
                                sequence += line;
                            }

                        }
                        if (firstLine != "")
                        {
                            ParseSputnikSSR(ssrs);
                            firstLine = "";
                        }
                        List<SSR> ssrList = new List<SSR>(bufferSize);
                        for (int i = 0; i < bufferSize; i++)
                        {
                            if (ssrs.Count > 0) ssrList.Add(ssrs.Dequeue());
                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRReaderModifiedSputnik2; {1} SSRs read in {2} ms", DateTime.Now.TimeOfDay, ssrList.Count, stopw.ElapsedMilliseconds));
#endif
                        return ssrList;

                    }

                    /// <summary>
                    /// The Modified Sputnik II file format does not support the SciRoKo SSR-search info
                    /// therefore this property returns false;
                    /// </summary>
                    public bool IsInfoReady
                    {
                        get
                        {
                            return false;
                        }
                    }

                    public OverallInfo ReadInfo()
                    {
                        OverallInfo info;
                        info.fileCount = 0;
                        info.nucleotideCount = 0;
                        return info;
                    }


                    //Extracts the SSR related information from the sputnikInfo string and stores the SSR in the database
                    private void ParseSputnikSSR(Queue<SSR> ssrs)
                    {
                        string[] division = this.firstLine.Split(new char[] { '\t' });
                        string motif = division[6];

                        SequenceCharacterKind motifKind = NucleotideSequenceUtility.GetSequenceCharacterKind(motif);

                        //Check if motif fits the criteria
                        if (motifKind != SequenceCharacterKind.ATCG) return;

                        int score = Convert.ToInt32(division[5]);

                        int ssrStart = Convert.ToInt32(division[2]);

                        int ssrEnd = Convert.ToInt32(division[3]);

                        int totalLength = Convert.ToInt32(division[4]);

                        division[7] += this.sequence;


                        //Also modified sputnik has the default mismatch penalty 6
                        //Determine number of mismatches
                        int mismatches = (totalLength - motif.Length - score) / 6;

                        //Add the Microsatellite
                        ssrs.Enqueue(new SSR(division[0], motif, SequenceFactory.GetSmallSequence(division[7]), ssrStart, ssrEnd, score + motif.Length, mismatches, SSRSearchMode.ModifiedSputnikII));
                        //Reset
                        this.firstLine = "";
                        this.sequence = "";
                    }

                }




                public class SSRWriterTabDelimited : ISSRWriter
                {
                    private StreamWriter sw;
                    public SSRWriterTabDelimited(StreamWriter sw)
                    {
                        this.sw = sw;

                    }
                    public void WriteSSRs(List<SSR> ssrs)
                    {

#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        sw.WriteLine("Seq_Name\tMotif\tMotif_Standardised\tSSR_Start\tSSR_End\tLength\tScore\tMismatches\tSearchMode\tSequence"); //9
                        foreach (SSR s in ssrs)
                        {
                            sw.WriteLine(String.Format("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}"
                                , s.ParentName.Replace('\t', ' '), s.Motif_Found, s.Motif_FullyStandardized, s.Start, s.End, s.Length, s.Score, s.Mismatches, s.SearchMode.ToString(), s.Sequence));
                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRWriterTabDelimited; {1} SSRs writen in TabDelimited style output format; Time used: {2} ms", DateTime.Now.TimeOfDay, ssrs.Count, stopw.ElapsedMilliseconds));

#endif

                    }

                    public void WriteInfo(OverallInfo searchInfo)
                    {
                        sw.WriteLine();
                        sw.WriteLine("#Nucleotide count: {0}", searchInfo.nucleotideCount);
                        sw.WriteLine("#File count: {0}", searchInfo.fileCount);
                        sw.WriteLine();
                    }

                }



                /// <summary>
                /// Reads SSRs from a TabDelimitedSSR-File
                /// </summary>
                public class SSRReaderTabDelimited : ISSRReader
                {
                    private StreamReader sr;
                    private readonly int bufferSize;
                    private bool infoReady = false;
                    private OverallInfo searchInfo;



                    public SSRReaderTabDelimited(StreamReader sr, int bufferSize)
                    {
                        this.sr = sr;
                        this.bufferSize = bufferSize;

                    }
                    public SSRReaderTabDelimited(StreamReader sr)
                    {
                        this.sr = sr;
                        this.bufferSize = 5000;

                    }

                    public List<SSR> ReadSSRs()
                    {
                        List<SSR> ssrList = new List<SSR>();
                        string line = "";


                        //sw.WriteLine("Seq_Name\tMotif\tMotif_Standardised\tSSR_Start\tSSR_End\tLength\tScore\tMismatches\tSequence"); //9

#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        //As long as there are lines in rhe file proceed!
                        while ((ssrList.Count < bufferSize) && ((line = sr.ReadLine()) != null))
                        {
                            //Get rid of space
                            line = line.Trim();
                            if (line == "") continue;


                            if (line.Contains("Seq_Name")) continue;
                            else if (line.StartsWith("#"))
                            {
                                int startPos = 0;
                                string substring = "";

                                if (line.Contains("Nucleotide"))
                                {
                                    startPos = line.IndexOf("count:") + 6;
                                    substring = line.Substring(startPos, line.Length - startPos);
                                    this.searchInfo.nucleotideCount = Convert.ToInt64(substring.Trim());
                                    infoReady = true;
                                }
                                else if (line.Contains("File"))
                                {
                                    startPos = line.IndexOf("count:") + 6;
                                    substring = line.Substring(startPos, line.Length - startPos);
                                    this.searchInfo.fileCount = Convert.ToInt32(substring.Trim());
                                    infoReady = true;


                                }

                            }
                            else
                            {
                                ssrList.Add(ParseTabDelimitedSSR(line));

                            }





                        }
#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRReaderTabDelimited; {1} SSRs read in {2} ms", DateTime.Now.TimeOfDay, ssrList.Count, stopw.ElapsedMilliseconds));
#endif

                        return ssrList;
                    }
                    /// <summary>
                    /// Indicates whether the Info is accessible and can be read;
                    /// SciRoKo usually stores the overall search info at the end of the file, therefore generally the info will be read after all
                    /// SSRs have been read
                    /// </summary>
                    public bool IsInfoReady
                    {
                        get
                        {
                            return infoReady;
                        }
                    }

                    /// <summary>
                    /// Returns an overall info containing the number of nucleotides and files that have been searched for microsatellites
                    /// </summary>
                    ///<exception cref="ArgumentException">The overall info is not yet ready; Use the IsInfoReady property prior to the use of the ReadInfo()</exception>
                    public OverallInfo ReadInfo()
                    {
                        if (!this.IsInfoReady) throw new ArgumentException("The overall info is not yet ready; Use the IsInfoReady property prior to the use of ReadInfo()");

                        return searchInfo;
                    }


                    //Extracts the SSR related information from the sputnikInfo string and stores the SSR in the database
                    private SSR ParseTabDelimitedSSR(string toParse)
                    {
                        //                  0       1           2               3          4       5       6        7           8           9
                        //sw.WriteLine("Seq_Name\tMotif\tMotif_Standardised\tSSR_Start\tSSR_End\tLength\tScore\tMismatches\tSearchMode\tSequence"); //9
                        string[] division = toParse.Split(new char[] { '\t' });
                        if (division.Length != 10) throw new Exception("Error during reading of Tab Delimited SSR-File.\nTabDelimited is not seccure, try the SciRoKo file format instead");
                        string motif = division[1];

                        int ssrStart = Convert.ToInt32(division[3]);
                        int ssrEnd = Convert.ToInt32(division[4]);
                        int length = Convert.ToInt32(division[5]);
                        int score = Convert.ToInt32(division[6]);
                        int mismatches = Convert.ToInt32(division[7]);


                        SSRSearchMode searchMode;
                        switch (division[8])
                        {
                            case "MisMatchFixedPenalty":
                                searchMode = SSRSearchMode.MisMatchFixedPenalty;
                                break;
                            case "MisMatchVariablePenalty":
                                searchMode = SSRSearchMode.MisMatchVariablePenalty;
                                break;
                            case "PerfectLength":
                                searchMode = SSRSearchMode.PerfectLength;
                                break;
                            case "PerfectRepeats":
                                searchMode = SSRSearchMode.PerfectRepeats;
                                break;
                            case "ModifiedSputnikI":
                                searchMode = SSRSearchMode.ModifiedSputnikI;
                                break;
                            case "ModifiedSputnikII":
                                searchMode = SSRSearchMode.ModifiedSputnikII;
                                break;
                            case "Sputnik":
                                searchMode = SSRSearchMode.Sputnik;
                                break;
                            case "PerfectMISA":
                                searchMode = SSRSearchMode.PerfectMISA;
                                break;
                            default: searchMode = SSRSearchMode.Unknown;
                                break;

                        }

                        //Add the Microsatellite
                        return new SSR(division[0], motif, SequenceFactory.GetSmallSequence(division[9]), ssrStart, ssrEnd, score, mismatches, searchMode);

                    }

                }


                /// <summary>
                /// Class for writing given lists of SSRs into a file according to the writing mode
                /// </summary>
                public class SSRWriterSciRoKo : ISSRWriter
                {
                    private StreamWriter sw;
                    public SSRWriterSciRoKo(StreamWriter sw)
                    {
                        this.sw = sw;
                    }

                    public void WriteInfo(OverallInfo searchInfo)
                    {
                        sw.WriteLine();
                        sw.WriteLine("#Nucleotide count: {0}", searchInfo.nucleotideCount);
                        sw.WriteLine("#File count: {0}", searchInfo.fileCount);
                        sw.WriteLine();

                    }


                    public void WriteSSRs(List<SSR> ssrs)
                    {
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        foreach (SSR s in ssrs)
                        {

                            string name = '>' + s.ParentName;
                            string info = string.Format("Motif: {0} - SSR start: {1} - SSR end: {2} - Total length: {3} - Score: {4} - Mismatches: {5} - Search Mode: {6}",
                                s.Motif_Found, s.Start, s.End, s.Length, s.Score, s.Mismatches, s.SearchMode.ToString());
                            string seq = s.Sequence.ToString();

                            sw.WriteLine(name);
                            sw.WriteLine(info);
                            sw.WriteLine(seq);



                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRWriterSciRoKo; {1} SSRs writen in SciRoKo-style output format; Time required: {2} ms", DateTime.Now.TimeOfDay, ssrs.Count, stopw.ElapsedMilliseconds));

#endif
                    }




                }

                /// <summary>
                /// Careful, Sputnik does not support mononucleotide or hexanucleotide ssr search
                /// </summary>
                public class SSRWriterSputnik : ISSRWriter
                {
                    private StreamWriter sw;
                    public SSRWriterSputnik(StreamWriter sw)
                    {
                        this.sw = sw;

                    }


                    public void WriteSSRs(List<SSR> ssrs)
                    {
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        foreach (SSR s in ssrs)
                        {
                            string prefix = "";
                            switch (s.MotifLength)
                            {
                                case 1: prefix = "single ";
                                    break;
                                case 2: prefix = "di";
                                    break;
                                case 3: prefix = "tri";
                                    break;
                                case 4: prefix = "tetra";
                                    break;
                                case 5: prefix = "penta";
                                    break;
                                case 6: prefix = "hexa";
                                    break;
                            }

                            string name = '>' + s.ParentName;

                            //The score in sputnik is calculated in a different way to SciRoKo therefore adjust it.
                            //Adjust: totalLength - motifLength - mm*mmP eg.: 40-3-3.6=21
                            string info = string.Format("{0}nucleotide {1} : {2} -- length {3} score {4}", prefix, s.Start, s.End, s.Length, s.Length - (s.MotifLength + s.Mismatches * 6));
                            string seq = s.Sequence.ToString();

                            sw.WriteLine(name);
                            sw.WriteLine(info);
                            sw.WriteLine(seq);

                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRWriterSputnik; {1} SSRs writen in Sputnik output format; Time used {2} ms", DateTime.Now.TimeOfDay, ssrs.Count, stopw.ElapsedMilliseconds));

#endif
                    }

                    public void WriteInfo(OverallInfo searchInfo)
                    {

                    }

                }

                /// <summary>
                /// Careful, Modified Sputnik I does not support hexanucleotide SSR search
                /// </summary>
                public class SSRWriterModifiedSputnik1 : ISSRWriter
                {
                    private StreamWriter sw;
                    public SSRWriterModifiedSputnik1(StreamWriter sw)
                    {
                        this.sw = sw;

                    }


                    public void WriteSSRs(List<SSR> ssrs)
                    {
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        foreach (SSR s in ssrs)
                        {
                            string prefix = "";
                            switch (s.MotifLength)
                            {
                                //carefull modified sputnik has mononucleotide instead of single nucleotide
                                case 1: prefix = "mono";
                                    break;
                                case 2: prefix = "di";
                                    break;
                                case 3: prefix = "tri";
                                    break;
                                case 4: prefix = "tetra";
                                    break;
                                case 5: prefix = "penta";
                                    break;
                                case 6: prefix = "hexa";
                                    break;
                            }

                            string name = '>' + s.ParentName;

                            //The score in sputnik is calculated in a different way to SciRoKo therefore adjust it.
                            //Adjust: totalLength - motifLength - mm*mmP eg.: 40-3-3.6=21
                            string info = string.Format("{0}nucleotide {1} : {2} -- length {3} score {4} unit {5}",
                                prefix, s.Start, s.End, s.Length, s.Score - s.MotifLength, s.Motif_FullyStandardized);
                            string seq = s.Sequence.ToString();

                            sw.WriteLine(name);
                            sw.WriteLine(info);
                            sw.WriteLine(seq);

                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRWriterModifiedSputnik1; {1} SSRs writen in Modified Sputnik I output format; Time used {2} ms", DateTime.Now.TimeOfDay, ssrs.Count, stopw.ElapsedMilliseconds));

#endif
                    }

                    public void WriteInfo(OverallInfo searchInfo)
                    {

                    }

                }

                /// <summary>
                /// Careful, Modified Sputnik II does not support hexanucleotide SSR search
                /// </summary>
                public class SSRWriterModifiedSputnik2 : ISSRWriter
                {
                    private StreamWriter sw;
                    public SSRWriterModifiedSputnik2(StreamWriter sw)
                    {
                        this.sw = sw;

                    }


                    public void WriteSSRs(List<SSR> ssrs)
                    {
#if DEBUG
                        System.Diagnostics.Stopwatch stopw = new System.Diagnostics.Stopwatch();
                        stopw.Start();
#endif
                        sw.WriteLine("#");
                        sw.WriteLine("# SEQID--REP_TYPE--SSRSTART--SSREND--SSRLEN--SCORE--MOTIF--SSR_SEQ");
                        sw.WriteLine("#");
                        sw.WriteLine();

                        foreach (SSR s in ssrs)
                        {
                            string prefix = "";
                            switch (s.MotifLength)
                            {
                                //carefull modified sputnik has mononucleotide instead of single nucleotide
                                case 1: prefix = "mono";
                                    break;
                                case 2: prefix = "di";
                                    break;
                                case 3: prefix = "tri";
                                    break;
                                case 4: prefix = "tetra";
                                    break;
                                case 5: prefix = "penta";
                                    break;
                                case 6: prefix = "hexa";
                                    break;
                            }



                            //The score in sputnik is calculated in a different way to SciRoKo therefore adjust it.
                            //Adjust: totalLength - motifLength - mm*mmP eg.: 40-3-3.6=21
                            string write = string.Format("{0}\t{1}nucleotide\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}",
                                s.ParentName, prefix, s.Start, s.End, s.Length, s.Score - s.MotifLength, s.Motif_FullyStandardized, s.Sequence);
                            sw.WriteLine(write);
                        }

#if DEBUG
                        stopw.Stop();
                        System.Diagnostics.Debug.WriteLine(String.Format("{0} SSRWriterModifiedSputnik2; {1} SSRs writen in Modified Sputnik II output format; Time used {2} ms", DateTime.Now.TimeOfDay, ssrs.Count, stopw.ElapsedMilliseconds));

#endif
                    }

                    public void WriteInfo(OverallInfo searchInfo)
                    {

                    }

                }



            }
        }


    }
}